/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file log.c
 *
 * @brief Implementation of logging functionality
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * @bugs No known bugs
 */

#include <sys/time.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "log.h"



/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

#define BUF_SIZE 4096



/* ====================== */
/* -- Static Variables -- */
/* ====================== */

static __thread int log_file;



/* ============================== */
/* -- Function Implementations -- */
/* ============================== */

/**
 * @brief Opens the log file for logging.
 *
 * If the log file does not exist, it is created. If it already exists, then it
 * is opened in append mode to continue logging.
 *
 * If no log file is passed, then the output is redirected to stderr.
 *
 * @param log_path Path to the log file
 */
void smr_log_open(const char *log_path)
{
    /* Open the log file */
    if (log_path != NULL) {
        if ((log_file = open(log_path, O_CREAT | O_RDWR, 0644)) < 0)
            exit(EXIT_FAILURE);
    } else
        log_file = fileno(stderr);
}



/**
 * @brief Closes the open log file.
 */
void smr_log_close()
{
    /* Attempt to close the log file */
    if (log_file != fileno(stderr))
        close(log_file);
}



/**
 * @brief Prints an error message to the log.
 *
 * @param function Name of the calling function
 * @param format Printf() style format string
 * @param ... Variable list of arguments
 */
void smr_log_msg(const char *function, const char *format, ...)
{
    va_list ap;
    char buf[BUF_SIZE];
    int len = 0;
    struct timeval timestamp;


    /* Read the variable argument list */
    va_start(ap, format);

    /* Construct the log message */
    if (gettimeofday(&timestamp, NULL)) {
        fprintf(stderr, "Error logging message: ");
        vfprintf(stderr, format, ap);
        fprintf(stderr, "\n");
        goto cleanup;
    }

    len = strftime(buf, BUF_SIZE, "%H:%M:%S.", localtime(&timestamp.tv_sec));
    len += snprintf(&buf[len], BUF_SIZE - len, "%6ld: ", timestamp.tv_usec);
    len += snprintf(&buf[len], BUF_SIZE - len, "%20s(): ", function);
    vsnprintf(&buf[len], BUF_SIZE - len, format, ap);
    len = strlen(buf);


    /* Output the message to the log file */
    if (write(log_file, buf, len) < len)
        fprintf(stderr, "Error logging message: %s\n", buf);


cleanup:
    /* Free the variable argument list */
    va_end(ap);
}



/**
 * @brief Prints the file attributes specified in the buffer passed
 *
 * @param function Name of the calling function
 * @param statbuf Buffer holding the file attributes to be logged
 */
void smr_log_stat(const char *function, const char *path,
                  const struct stat *statbuf) {
    char buf[BUF_SIZE];
    int len = 0;


    /* Construct the log message using the fields of stat buffer */
    len += snprintf(&buf[len], BUF_SIZE - len, "%s:"
                    "st_dev: %ld, "
                    "st_ino: %ld, "
                    "st_mode: 0%o, "
                    "st_nlink: %ld, "
                    "st_uid: %d, "
                    "st_gid: %d, "
                    "st_rdev: %ld, "
                    "st_size: %ld, "
                    "st_blksize: %ld, "
                    "st_blocks: %ld, "
                    "st_atime: 0x%8lx, "
                    "st_mtime: 0x%8lx, "
                    "st_ctime: 0x%8lx\n",
                    path, statbuf->st_dev, statbuf->st_ino, statbuf->st_mode,
                    statbuf->st_nlink, statbuf->st_uid, statbuf->st_gid,
                    statbuf->st_rdev, statbuf->st_size, statbuf->st_blksize,
                    statbuf->st_blocks, statbuf->st_atime, statbuf->st_mtime,
                    statbuf->st_ctime);

    /* Output the message to the log file */
    smr_log_msg(function, "%s", buf);
}
