/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file emudiskimg.c
 *
 * @brief Defines the Macros, data types and functions for using the SMR
 * Emulator.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#define _XOPEN_SOURCE 500

#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include "log.h"
#include "debug.h"

#include "emulator.h"
#include "emudiskimg.h"
#include "ediversion.h"



/* -- Macro Definitions -- */

#define EDIOPS(fd)      edi_versions[edis[fd].version]

/**
 * @brief Magic Number identifying the disk image files.
 */
#define EDI_MAGIC_NUMBER 0xE5D10000U     // Hex for Emulated 5hingled Disk 1mage



/* -- Type Definitions -- */

/**
 * @brief Structure of the file descriptor lookup table used by the Emulator.
 *
 * @param fd Actual file descriptor of an open EDI file
 * @param version The EDI version implemented in the file
 */
typedef struct {
    int fd;
    uint32_t version;
} edi_fd_t;



/* -- Static Variables -- */
static edi_fd_t edis[MAX_EDIS];
static int edi_count;



/* -- Function Implementations -- */

/**
 * @brief Creates a new EDI file based on the parameters specified.
 *
 * @param filename The name of the new EDI file
 * @param version The EDI format version to use for creating the new EDI file
 * @param ... Version specific parameters
 *
 * @return On success, the file descriptor of the newly created EDI file;
 *                 -1 on failure
 *
 * Error Codes:
 * - EVERSION Invalid/Unknown version number
 * - EUNSUPPORTED The version is not supported
 *
 * Other than the above, the function also returns the error codes returned by
 * open(), pwrite() and the create() function for the specific EDI version.
 */
#ifdef _EDI_ON_RAW
int edi_create(const char* metadata_filename, const char* filename, uint32_t version, ...) {
#else
int edi_create(const char* filename, uint32_t version, ...) {
#endif
    int fd, retval = -1;
    int oflags = O_RDWR | O_CREAT | O_EXCL;
    mode_t mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
    edi_ident_t edi_ident;
    va_list args;


    /* Validate the version */
    if (!ediv_is_valid(version)) {
        errno = EVERSION;
        return -1;
    }


    /* Check if the version specific create() function exists */
    if (edi_versions[version]->create == NULL) {
        errno = EUNSUPPORTED;
        return -1;
    }


    /* Process the variable arguments list */
    va_start(args, version);


    /* Create the EDI file */
#ifdef _EDI_ON_RAW
    if ((fd = open(metadata_filename, oflags, mode)) < 0) {
#else
    if ((fd = open(filename, oflags, mode)) < 0) {
#endif
	perror("edi_create");
        goto cleanup_vargs;
    }


    /* Write the EDI identifier record */
    edi_ident.magic_no = EDI_MAGIC_NUMBER;
    edi_ident.version = version;
    if (pwrite(fd, &edi_ident, sizeof(edi_ident_t), 0) != sizeof(edi_ident_t))
        goto cleanup_file;


    /* Call the version specific create() function to handle further tasks */
    if (edi_versions[version]->create(fd, args) < 0)
        goto cleanup_file;
    dbg("Header and Metadata written out to file!\n");

    /* Return the file descriptor to the caller */
    retval = fd;
    goto cleanup_vargs;


cleanup_file:
    /* Close and delete the open file */
    close(fd);
    unlink(filename);

cleanup_vargs:
    /* Cleanup the variable arguments */
    va_end(args);

    return retval;
}



/**
 * @brief Dumps the metadata for the specified EDI file
 *
 * Calling this function sends output to stdout. All metadata for the specified
 * EDI file is printed, including the version, block size, band count, and the
 * write pointers for each band.
 *
 * @param fd The file descriptor index returned by edi_open()
 *
 * @return Depends on the version
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 * - EUNSUPPORTED - The version of the EDI does not support the operation
 */
int edi_dump(int fd) {
    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        errno = EBADFD;
        return -1;
    }

    /* Select the appropriate call-back function */
    if (EDIOPS(fd)->dump == NULL) {
        dbg("version[%d]->dump() not supported!\n", edis[fd].version);
        errno = EUNSUPPORTED;
        return -1;
    }

    /* Call the registered call-back function */
    dbg("Calling version[%d]->dump()!\n", edis[fd].version);
    return EDIOPS(fd)->dump(edis[fd].fd);
}



/**
 * @brief Opens the specified EDI file and verifies its validity.
 *
 * In case or any errors, the function DOES NOT close the open EDI file. It is
 * up to the programmer to detect any errors, handle them accordingly and then
 * close the open EDI file.
 *
 * @param filename Path to the EDI file.
 *
 * @return On success, the index of the file descriptor in the edis array;
 *                 -1 on failure
 *
 * Errors returned:
 * - EINVALIDFILE The file is not a valid EDI file
 * - EVERSION The version of the EDI file is unknown
 * - EUNSUPPORTED The version of the EDI file is unsupported
 *
 * Other than the above, the function also returns the error codes returned by
 * open() and the open() function for the specific EDI version.
 */
#ifdef _EDI_ON_RAW
int edi_open(const char* metadata_filename, const char* filename, int * edi_meta_fd) {
#else
int edi_open(const char* filename) {
#endif
    /* Declare the variables needed */
    int edi_fd;
    edi_ident_t edi_ident;
    int ident_size = sizeof(edi_ident_t);

    int flags = O_RDWR;
#ifdef _EDI_ON_RAW
    if ((*edi_meta_fd = open(metadata_filename, flags)) < 0) {
	return -1;
    }

    flags |= O_DIRECT;
#endif

    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_open(filename=%s)\n", filename);


    /* Attempt to open the file for reading and writing */
    if ((edi_fd = open(filename, flags)) < 0) {
        return -1;
    }


    /* Read the header to check if the file is a valid Emulated Disk Image */
#ifdef _EDI_ON_RAW
    if (pread(*edi_meta_fd, &edi_ident, ident_size, 0) < ident_size) {
#else
    if (pread(edi_fd, &edi_ident, ident_size, 0) < ident_size) {
#endif
        /* Report the error */
        log(LOG_EMU_ERRORS, "Cannot read identifier data!\n");
        errno = EINVALIDFILE;
        return -1;
    }

    /* Verify the Magic Number */
    if (edi_ident.magic_no != EDI_MAGIC_NUMBER) {
        /* Report the error */
	log(LOG_EMU_ERRORS, "Magic Number does not match!\n");
        errno = EINVALIDFILE;
        return -1;
    }


    /* Verify that the version is valid and supported */
    if (!ediv_is_valid(edi_ident.version)) {
        log(LOG_EMU_ERRORS, "Version is not valid!\n");
        errno = EVERSION;
        return -1;
    }

    if (edi_versions[edi_ident.version] == NULL) {
        log(LOG_EMU_ERRORS, "Version %d is not supported by the library!\n",
            edi_ident.version);
        errno = EUNSUPPORTED;
        return -1;
    }


    /* Add the new file descriptor to the list of open EDI files */
    edis[edi_count].fd = edi_fd;
    edis[edi_count].version = edi_ident.version;

#ifdef _EDI_ON_RAW
    edi_count++;
    edis[edi_count].fd = *edi_meta_fd;
    edis[edi_count].version = edi_ident.version;
    *edi_meta_fd = edi_count;
#endif

    /* Call the registered callback function, if any */
    if (EDIOPS(edi_count)->open != NULL) {
        if ((EDIOPS(edi_count)->open(edis[edi_count].fd)) < 0)
            return -1;
    }


    /* Return the index of the newly opened EDI in the open_edis array */
    log(LOG_EMU_RETURN, "EDI File %s opened with descriptor %d!\n",
        filename, edis[edi_count].fd);
    edi_count++;

#ifdef _EDI_ON_RAW
    return (edi_count - 2);
#else
    return (edi_count - 1);
#endif
}



/**
 * @brief Closes an open EDI file
 *
 * @param fd The file descriptor index returned by edi_open()
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 *
 * Other than the above, the function also returns the error codes returned by
 * close() and the close() function call for the specific EDI version.
 */
#ifdef _EDI_ON_RAW
int edi_close(int meta_fd, int fd) {
#else
int edi_close(int fd) {
#endif
    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_close(fd=%d)\n", fd);


    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        log(LOG_EMU_ERRORS, "Invalid index %d into the FD lookup table!\n", fd);
        errno = EBADFD;
        return -1;
    }


    /* Call the registered callback function, if any */
    if (EDIOPS(fd)->close != NULL) {
        EDIOPS(fd)->close(edis[fd].fd);
    }


    /* Attempt to close the file */
    if (close(edis[fd].fd) < 0) {
        log(LOG_EMU_ERRORS, "Cannot close EDI file!\n");
        return -1;
    }

#ifdef _EDI_ON_RAW
    if (close(meta_fd) < 0) {
	log(LOG_EMU_ERRORS, "Cannot close metadata file!\n");
	return -1;
    }
#endif

    log(LOG_EMU_RETURN, "EDI file closed successfully!\n");
    return 0;
}



/**
 * @brief Queries the Emulated Disk to retrieve disk parameters such as block
 * size, band count, etc.
 *
 * This function takes a void pointer for the data variable. However,
 * internally, it will cast the pointer to the desired type depending on the
 * data being requested. It is up to the programmer to ensure that the pointer
 * being passed addresses the required amount of memory for the returned data.
 *
 * @see inq_t
 *
 * @param fd The file descriptor index returned by edi_open()
 * @param inq The Inquiry code for the desired data
 * @param data Pointer to a variable to store the returned data
 *
 * @return Depends on the version
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 * - EUNSUPPORTED - The version of the EDI does not support the operation
 */
int edi_modesense(int fd, inq_t inq, void *data) {

    /* Log the function call */
    switch(inq) {
    case INQ_BLKSIZE:
        log(LOG_EMU_CALLS, "edi_modesense(fd=%d, inq=BLKSIZE, data=%p)\n",
            fd, data);
        break;

    case INQ_BNDSIZE:
        log(LOG_EMU_CALLS, "edi_modesense(fd=%d, inq=BNDSIZE, data=%p)\n",
            fd, data);
        break;

    case INQ_BNDCOUNT:
        log(LOG_EMU_CALLS, "edi_modesense(fd=%d, inq=BANDCOUNT, data=%p)\n",
            fd, data);
        break;

    }


    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        log(LOG_EMU_ERRORS, "Invalid index %d into the FD lookup table!\n", fd);
        errno = EBADFD;
        return -1;
    }

    /* Select the appropriate call-back function */
    if (EDIOPS(fd)->modesense == NULL) {
        log(LOG_EMU_ERRORS, "version[%d]->modesense() not supported!\n",
            edis[fd].version);
        errno = EUNSUPPORTED;
        return -1;
    }

    /* Call the registered call-back function */
    return EDIOPS(fd)->modesense(edis[fd].fd, inq, data);
}



/**
 * @brief Manages bands on the Emulated Disk.
 *
 * The Manage Bands command is used for multiple actions, depending on the
 * Service Action requested. Each Service Action expects a certain number of
 * parameters for the function to execute correctly.
 *
 * 1. SVC_GETPTR: Retrieves the Write Pointer for the specified band. It
 * expects 2 parameters:
 * - The Band No. of the band whose Write Pointer is to be retrieved
 * - A pointer to a variable to store the retrieved RBA.
 *
 * 2. SVC_SETPTR: Sets the Write Pointer for the specified band. It expects 2
 * parameters:
 * - The Band No. of the band whose Write Pointer is to be set
 * - The new RBA to set as the Write Pointer for the band
 *
 * 3. SVC_GETMULTIPLE: Retrieves the Write Pointers for multiple bands. It
 * expects 3 parameters:
 * - An array of Band Numbers whose Write Pointers are to be retrieved.
 * - An array to store the Write Pointers of the bands
 * - A count of the number of bands passed in the array
 *
 * 4. SVC_RESETALL: Resets the Write Pointers of all bands to the start of the
 * band. This function does not expect any parameters.
 *
 * 5. SVC_GETBANDCONFIG: Retrieves the band configuration for the specific band.
 * It expects 2 parameters:
 * - The Band No. of the band whose configuration is to be retrieved
 * - A pointer to a variable to store the retrieved configuration.
 *
 *
 * @param fd The file descriptor index returned by edi_open()
 * @param action The desired action to perform
 * @param band Pointer to the band on which the action is to be performed
 * @param rba Pointer to the RBA for the operation
 * @param len Length of the array of band numbers (only for SVC_GETMULTIPLE)
 *
 * @return Depends on the version
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 * - EUNSUPPORTED - If the version of the EDI does not support the operation
 *
 * Other than the above, the function also returns the error codes returned by
 * the managegbands() function call for the specific EDI version.
 */
int edi_managebands(int fd, svc_act_t action, band_t* band, rba_t* rba,
                    int len) {
    /* Log the function call */
    switch(action) {
    case SVC_GETPTR:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_GETPTR, band=%d, "
            "rba=%p)\n", fd, *band, rba, len);
        break;

    case SVC_SETPTR:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_SETPTR, band=%d, "
            "rba=%d)\n", fd, *band, *rba, len);
        break;

    case SVC_GETMULTIPLE:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_GETMULTIPLE, band=%p"
            ", rba=%p, len=%d)\n", fd, band, rba, len);
        break;

    case SVC_RESETALL:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_RESETALL, band=%p, "
            "rba=%p, len=%d)\n", fd, band, rba, len);
        break;

    case SVC_GETBANDCONFIG:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_GETBANDCONFIG, "
            "band=%d, rba=%d, len=%d)\n", fd, *band, *rba, len);
        break;

    case SVC_SETBANDCONFIG:
        log(LOG_EMU_CALLS, "edi_managebands(fd=%d, inq=SVC_SETBANDCONFIG, "
            "band=%d, rba=%d, len=%d)\n", fd, *band, *rba, len);
        break;
    }


    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        log(LOG_EMU_ERRORS, "Invalid index %d into the FD lookup table!\n", fd);
        errno = EBADFD;
        return -1;
    }

    /* Select the appropriate call-back function */
    if (EDIOPS(fd)->managebands == NULL) {
        log(LOG_EMU_CALLS, "version[%d]->modesense() not supported!\n",
            edis[fd].version);
        errno = EUNSUPPORTED;
        return -1;
    }

    /* Call the registered call-back function */
    return EDIOPS(fd)->managebands(edis[fd].fd, action, band, rba, len);
}



/**
 * @brief Reads data from the specified band and block.
 *
 * It is up to the programmer to ensure that the buffer passed to the function
 * is sufficiently large to hold the requested data.
 *
 * @param fd The file descriptor index returned by edi_open()
 * @param band The Band no. of the band from which data is to be read
 * @param rba The starting block in the band from which data is to be read
 * @param buf Pointer to the buffer into which the data is to be read
 * @param len Number of contiguous "blocks" to read, starting from rba
 *
 * @return Depends on the version
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 * - EUNSUPPORTED - If the version of the EDI does not support the operation
 */
#ifdef _EDI_ON_RAW
int edi_read(int meta_fd, int fd, band_t band, rba_t rba, char *buf, rba_t len) {
    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_read(meta_fd=%d, fd=%d, band=%d, rba=%d, buf=%p, len=%d)\n",
       meta_fd, fd, band, rba, buf, len);
#else
int edi_read(int fd, band_t band, rba_t rba, char *buf, rba_t len) {
    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_read(fd=%d, band=%d, rba=%d, buf=%p, len=%d)\n",
       fd, band, rba, buf, len);
#endif

    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        log(LOG_EMU_ERRORS, "Invalid index %d into the FD lookup table!\n", fd);
        errno = EBADFD;
        return -1;
    }

    /* Select the appropriate call-back function */
    if (EDIOPS(fd)->read == NULL) {
        log(LOG_EMU_ERRORS, "version[%d]->read() not supported!\n",
            edis[fd].version);
        errno = EUNSUPPORTED;
        return -1;
    }

    /* Call the registered call-back function */
#ifdef _EDI_ON_RAW
    return EDIOPS(fd)->read(edis[meta_fd].fd, edis[fd].fd, band, rba, buf, len);
#else
    return EDIOPS(fd)->read(edis[fd].fd, band, rba, buf, len);
#endif
}



/**
 * @brief Writes data to the specified band and block.
 *
 * @param fd The file descriptor index returned by edi_open()
 * @param band The Band no. of the band to which data is to be written
 * @param rba The starting block in the band to which data is to be written
 * @param buf Pointer to the buffer which holds the data is to be written
 * @param len Number of contiguous "blocks" to write, starting from rba
 *
 * @return Depends on the version
 *
 * Error Codes:
 * - EBADFD - The file descriptor index passed is invalid
 * - EUNSUPPORTED - If the version of the EDI does not support the operation
 */
#ifdef _EDI_ON_RAW
int edi_write(int meta_fd, int fd, band_t band, rba_t rba, char *buf, rba_t len) {
    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_write(meta_fd = %d, fd=%d, band=%d, rba=%d, buf=%p, len=%d)\n",
        meta_fd, fd, band, rba, buf, len);
#else
int edi_write(int fd, band_t band, rba_t rba, char *buf, rba_t len) {
	    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_write(fd=%d, band=%d, rba=%d, buf=%p, len=%d)\n",
        fd, band, rba, buf, len);
#endif

    /* Verify the file descriptor index */
    if (fd >= edi_count) {
        log(LOG_EMU_ERRORS, "Invalid index %d into the FD lookup table!\n", fd);
        errno = EBADFD;
        return -1;
    }

    /* Select the appropriate call-back function */
    if (EDIOPS(fd)->write == NULL) {
        log(LOG_EMU_ERRORS, "version[%d]->write() not supported!\n",
            edis[fd].version);
        errno = EUNSUPPORTED;
        return -1;
    }

    /* Call the registered callback function */
#ifdef _EDI_ON_RAW
    return EDIOPS(fd)->write(edis[meta_fd].fd, edis[fd].fd, band, rba, buf, len);
#else
    return EDIOPS(fd)->write(edis[fd].fd, band, rba, buf, len);
#endif
}



/**
 * Initializes the Emulated Disk Image interface.
 *
 * @return 0 on success; -1 on failure
 */
int edi_init() {
    /* Log the function call */
    log(LOG_EMU_CALLS, "edi_init()\n");


    /* Initialize the data structures */
    memset(&edis, 0, sizeof(edi_fd_t) * MAX_EDIS);
    edi_count = 0;


    /* Call the initialization operations for the Versioning Interface */
    return ediv_init();
}
