/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file ediv0.c
 *
 * @brief Implementation of functionality for version 0 of the Emulator Disk
 * Image format.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * TODO Document errors for create()
 */

#define _XOPEN_SOURCE 500

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "log.h"
#include "debug.h"
#include "ediversion.h"



/* -- Macro Definitions -- */
#define EDIVERSION 0

#define HEADER		ediv0_header

#define OFF_META	OFF_HEADER + sizeof(ediv0_header_t)

#ifdef _EDI_ON_RAW
#define OFF_DATA        0 /* Data will start at the beginning of the raw device */
#else
#define OFF_DATA    OFF_META + sizeof(ediv0_metadata_t) * HEADER.band_count
#endif


/* -- Type Definitions -- */

/**
 * @brief Defines the header for the Emulated Disk Image (EDI).
 *
 * Version 1 of the EDI defines an emulated disk of fixed size bands. The
 * header file defines the size of a block on disk, the total number of
 * shingled bands on the disk and the size of a band, in terms of the number
 * of blocks it is made up of.
 *
 * @param blk_size Size of a disk block, in bytes;
 * @param band_count Number of bands on the emulated disk
 * @param band_size Size of a band, in terms of the number of blocks it spans.
 */
typedef struct {
    uint32_t blk_size;
    band_t band_count;
    rba_t band_size;
} ediv0_header_t;


/**
 * @brief Defines the structure of a metadata record.
 *
 * The structure of a metadata record defines the size of the band and the
 * current position of the write pointer (cursor). The band number serves as
 * an index to find the record for a band.
 *
 * @param size Size of the band, in terms of the number of blocks it spans
 * @param write_ptr The last written block in the band
 */
typedef struct {
    rba_t write_ptr;
} ediv0_metadata_t;



/* -- Static Variables -- */
static edi_vops_t ediv0_ops;
static ediv0_header_t ediv0_header;



/* -- Function Prototypes -- */
static int is_valid_band(band_t);
static int get_meta_rec(int, band_t, ediv0_metadata_t*);
static int set_meta_rec(int, band_t, ediv0_metadata_t*);


/* -- Function Implementations -- */

/**
 * @brief Writes out the EDI data structures for the v0 EDI file format.
 *
 * @param fd File descriptor of the open EDI file
 * @param args List of arguments describing v0 EDI parameters
 *
 * @return 0 on success; -1 on failure
 */
static int ediv0_create(int fd, va_list args) {
	ediv0_header_t header;
	ediv0_metadata_t metadata;
	size_t size;
    uint32_t blk_size;
    band_t band_count;
    rba_t band_size;
    int i;


	/* Verify if all the required arguments are present and valid */
    blk_size = va_arg(args, uint32_t);
    band_count = va_arg(args, band_t);
    band_size = va_arg(args, rba_t);

    if ((blk_size & (blk_size - 1)))	/* Block size must be a power of 2 */
    	return -1;

    if (band_count <= 0)				/* Band count must be greater than 0 */
    	return -1;

    if (band_size <= 0)					/* Band size must be greater than 0 */
       	return -1;


    /* Write out the EDI header */
    header.blk_size = blk_size;
    header.band_count = band_count;
    header.band_size = band_size;

    size = sizeof(ediv0_header_t);
    if (pwrite(fd, &header, size, OFF_HEADER) != size)
    	return -1;


    /* Write out the metadata */
    metadata.write_ptr = 0;
    size = sizeof(ediv0_metadata_t);

    if (lseek(fd, 0, SEEK_END) < 0) {
    	dbg("Unable to see to end of file!\n");
    	return -1;
    }

    for(i=0; i<band_count; i++) {
        if (write(fd, &metadata, size) != size) {
            dbg("Metadata record %d failed to be written out!\n", i);
        	return -1;
        }
    }

	return 0;
}



/**
 * @brief Writes out the EDI data structures for the v0 EDI file format.
 *
 * @param fd File descriptor of the open EDI file
 * @param args List of arguments describing v0 EDI parameters
 *
 * @return 0 on success; -1 on failure
 */
static int ediv0_dump(int fd) {
	ediv0_metadata_t metadata;
    int i;
    int cols = 4;


    /* Display the EDI Version and basics */
    fprintf(stdout, "EDI INFORMATION\n===============\n");
    fprintf(stdout, "EDI Version: 0\n");
    fprintf(stdout, "Block Size: %d bytes\n", ediv0_header.blk_size);
    fprintf(stdout, "Band Size: %d\n", ediv0_header.band_size);
    fprintf(stdout, "No. of bands: %d\n", ediv0_header.band_count);
    fprintf(stdout, "Write Pointers for bands:");


    /* Display the write pointers for each band */
    for (i=0; i<ediv0_header.band_count; i++) {
    	/* Display the row header */
    	if (!(i % cols))
    		fprintf(stdout, "\nBand %4d:", i);

    	/* Read the band meta data for the current band */
    	if (get_meta_rec(fd, i, &metadata) < 0) {
    		log(LOG_EMU_ERRORS, "Error reading metadata for band %d: %s\n",
    							i, strerror(errno));
    		return -1;
    	}
    	fprintf(stdout, "    %8d", metadata.write_ptr);
    }
    fprintf(stdout, "\n\n");

	return 0;
}



/**
 * @brief Loads the Header for the v0 EDI on open().
 *
 * @param fd File descriptor for the open EDI file.
 *
 * @return 0 on success; -1 on failure
 */
static int ediv0_open(int fd) {
	int size;


	/* Read the header data from disk */
	size = sizeof(ediv0_header_t);
	if (pread(fd, &ediv0_header, size, OFF_HEADER) < size) {
    	log(LOG_EMU_ERRORS, "Cannot read v0 header!\n");
		errno = EFAIL;
		return -1;
	}

	return 0;
}



/**
 * @brief Implements the modesense() function.
 *
 * Supported Operations:
 * - INQ_BLKSIZE - Returns the block size for the EDI
 * - INQ_BNDCOUNT - Returns the number of bands on the EDI
 *
 * @param fd File descriptor of the open EDI file
 * @param inq The type of inquiry
 * @param data Pointer to the buffer to store the data returned
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - EFAIL - Failed to read the header from the EDI file
 * - EUNSUPPORTED - Unsupported inquiry type
 */
static int ediv0_modesense(int fd, inq_t inq, void* data) {
	/* Check the operation requested */
	switch(inq) {
	case INQ_BLKSIZE:		/* Block Size */
		*((uint32_t *)data) = ediv0_header.blk_size;
		return 0;

	case INQ_BNDSIZE:		/* Band Size */
		*((rba_t *)data) = ediv0_header.band_size;
		return 0;

	case INQ_BNDCOUNT:		/* Band Count */
		*((band_t *)data) = ediv0_header.band_count;
		return 0;

	default:				/* Unsupported Operation */
		errno = EUNSUPPORTED;
		fprintf(stderr,"\treturning through unsupported with fd %d\n", fd);
		return -1;
	}
}



/**
 * @brief Implements the managebands() function.
 *
 * This version only implements the GETPTR, SETPTR, GETMULTIPLE, RESETALL and
 * GETBANDCONFIG operations. All other operations are unsupported.
 *
 * @param fd File descriptor of the open EDI file
 * @param action The action to be performed
 * @param band Pointer to the band to be passed to the operation
 * @param rba Pointer to the RBA to be passed to the operation
 * @param len Length of the array of band numbers. Only used for GETMULTIPLE
 *
 * @return 0 on success; -1 on failure
 */
static int ediv0_managebands(int fd, svc_act_t action, band_t* band, rba_t* rba,
		int len) {
	ediv0_metadata_t metadata;
	int i, rba_zero = 0, retval = 0;


	/* Check the action requested */
	switch (action) {
	case SVC_GETPTR:		/* Get Write Pointer for band */
		/* Read the metadata for the band */
		if (get_meta_rec(fd, *band, &metadata) < 0)
			return -1;

		/* Return the Write Pointer */
		*rba = metadata.write_ptr;
		return 0;


	case SVC_SETPTR:		/* Set Write Pointer for band */
		/* Verify that the rba is valid */
		if ((*rba < 0) || (*rba >= ediv0_header.band_size)) {
			log(LOG_EMU_ERRORS, "Invalid RBA (%d) specified!", *rba);
			errno = ERBA;
			return -1;
		}

		/* Read the metadata for the band */
		if (get_meta_rec(fd, *band, &metadata) < 0)
			return -1;

		/* Update the Write Pointer */
		metadata.write_ptr = *rba;

		/* Write the metadata for the band */
		if (set_meta_rec(fd, *band, &metadata) < 0)
			return -1;

		return 0;


	case SVC_GETMULTIPLE:	/* Get Write Pointer for bands */
		/* Loop over all the bands passed */
		for(i=0; i<len; i++) {
			/* Retrieve the Write Pointer for the current band */
			if (ediv0_managebands(fd, SVC_GETPTR, &band[i], &rba[i], 0) < 0) {
				rba[i] = -1;
				retval = -1;
				log(LOG_EMU_ERRORS, "Unable to retrieve Write Pointer for band "
									"%d\n", band[i]);
			}
		}

		return retval;


	case SVC_RESETALL:		/* Set Write Pointers for all bands */
		/* Loop over all the bands */
		for(i=0; i<ediv0_header.band_count; i++) {
			/* Retrieve the Write Pointer for the current band */
			if (ediv0_managebands(fd, SVC_SETPTR, &band[i], &rba_zero, 0) < 0) {
				rba[i] = -1;
				retval = -1;
				log(LOG_EMU_ERRORS, "Unable to retrieve Write Pointer for band "
									"%d\n", band[i]);
			}
		}

		return 0;


	case SVC_GETBANDCONFIG:	/* Returns the size of a band */
		/* Validate the band */
		if (!is_valid_band(*band)) {
			log(LOG_EMU_ERRORS, "Band %d is not valid!\n", *band);
			errno = EBAND;
			return -1;
		}

		/* Return the size of the band */
		*rba = ediv0_header.band_size;
		return 0;


	default:				/* Unsupported Operation */
		log(LOG_EMU_ERRORS, "Unsupported operation %x requested for band %d\n",
							action, *rba);
		errno = EUNSUPPORTED;
		return -1;
	}
}



/**
 * @brief Implements the read() function.
 *
 * @param fd File descriptor of the open EDI file
 * @param band The band from which data is to be read
 * @param rba The rba from which the read is to be started
 * @param buf Pointer to the buffer to hold the data read
 * @param count The number of blocks to read
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - ERBA - The RBA passed is negative or greater than the band size
 * - EOOB - The RBA passed is greater than the Write Pointer for the band or the
 * 			operation will go past the Write Pointer for the count specified.
 *
 * Other than the above, the function also returns the error codes returned by
 * pread() and get_meta_rec() for the current version.
 */
#ifdef _EDI_ON_RAW
static int ediv0_read(int meta_fd, int fd, band_t band, rba_t rba, char* buf, rba_t count) {
#else
static int ediv0_read(int fd, band_t band, rba_t rba, char* buf, rba_t count) {
#endif
	ediv0_metadata_t metadata;
	off_t offset;
	size_t size = count * HEADER.blk_size;


	/* Read the metadata for the band */
#ifdef _EDI_ON_RAW
	if (get_meta_rec(meta_fd, band, &metadata) < 0) {
#else
	if (get_meta_rec(fd, band, &metadata) < 0) {
#endif
		return -1;
	}

	/* Verify that the rba is valid */
	if ((rba < 0) || (rba >= HEADER.band_size)) {
		errno = ERBA;
		return -1;
	}

	if (rba >= metadata.write_ptr) {
		errno = EOOB;
		return -1;
	}

	/* Check the validity of the read length */
	if (rba + count > metadata.write_ptr) {
		errno = EOOB;
		return -1;
	}


	/* Read the data into the buffer */
	offset = OFF_DATA + ((off_t)band * HEADER.band_size * HEADER.blk_size);
	offset += ((off_t)rba * HEADER.blk_size);
	if (pread(fd, buf, size, offset) != size)
		return -1;

	return 0;
}



/**
 * @brief Implements the write() function.
 *
 * @param fd File descriptor of the open EDI file
 * @param band The band from which data is to be written
 * @param rba The rba from which the write should begin. Should match the Write
 * 			  Pointer for the band.
 * @param buf Pointer to the buffer holding the data to be written
 * @param count The number of blocks to write
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - ERBA - The RBA passed does not match the Write Pointer for the band
 * - EOOB - The RBA passed is greater than the Write Pointer for the band or the
 * 			operation will go past the Write Pointer for the count specified.
 * - EUPDATE - There was an error updating the Write Pointer for the band
 *
 * Other than the above, the function also returns the error codes returned by
 * pread() and get_meta_rec() for the current version.
 */
#ifdef _EDI_ON_RAW
static int ediv0_write(int meta_fd, int fd, band_t band, rba_t rba, char* buf, rba_t count) {
#else
static int ediv0_write(int fd, band_t band, rba_t rba, char* buf, rba_t count) {	
#endif
	ediv0_metadata_t metadata;
	off_t offset;
	off_t size = count * HEADER.blk_size;


	/* Check the validity of the write length */
	if ((rba + count > rba) && (rba + count > ediv0_header.band_size)) {
		log(LOG_EMU_ERRORS, "rba+count (%d) exceed band size (%d)!\n",
							rba+count, ediv0_header.band_size);
		errno = EOOB;
		return -1;
	}

	/* Read the metadata for the band */
#ifdef _EDI_ON_RAW
	if (get_meta_rec(meta_fd, band, &metadata) < 0) {
#else
	if (get_meta_rec(fd, band, &metadata) < 0) {
#endif
		log(LOG_EMU_ERRORS, "Cannot fetch metadata record for band %d\n", band);
		return -1;
	}

	/* Verify that the rba is valid */
	if (rba != metadata.write_ptr) {
		log(LOG_EMU_ERRORS, "rba (%d) does not match write pointer (%d)\n",
				rba, metadata.write_ptr);
		errno = ERBA;
		return -1;
	}

	/* Write the data from the buffer */
	offset = OFF_DATA + ((off_t)band * HEADER.band_size * HEADER.blk_size);
	offset += ((off_t)rba * HEADER.blk_size);
	if (pwrite(fd, buf, size, offset) != size) {
		log(LOG_EMU_ERRORS, "pwrite(): %s\n", strerror(errno));
		return -1;
	}

	/* Update the write pointer for the band */
	metadata.write_ptr += count;
#ifdef _EDI_ON_RAW
	if (set_meta_rec(meta_fd, band, &metadata) < 0) {
#else
	if (set_meta_rec(fd, band, &metadata) < 0) {
#endif
		log(LOG_EMU_ERRORS, "Failed to update write pointer (%d)\n",
							metadata.write_ptr);
		errno = EUPDATE;
		return -1;
	}

	return 0;
}



/**
 * @brief Initializes the EDIv0 code base.
 *
 * @return Pointer to the EDIv0 Call-back Function Set
 */
edi_vops_t* ediv0_init() {
	/* Initialize all data structures */
	memset(&ediv0_header, 0, sizeof(ediv0_header_t));


	/* Initialize the Call-back Functions object and return it to the caller */
	ediv0_ops.create = ediv0_create;
	ediv0_ops.dump = ediv0_dump;
	ediv0_ops.open = ediv0_open;
	ediv0_ops.modesense = ediv0_modesense;
	ediv0_ops.managebands = ediv0_managebands;
	ediv0_ops.read = ediv0_read;
	ediv0_ops.write = ediv0_write;

	return &ediv0_ops;
}



/**
 * @brief Returns whether or not the specified band is valid for the current EDI
 *
 * @param band The band number to be verified
 *
 * @return 1 on success; 0 on failure
 */
static int is_valid_band(band_t band) {
	/* Ensure that the band number is less than the band count for the EDI */
	return ((band >= 0) && (band < ediv0_header.band_count));
}



/**
 * @brief Returns the Metadata record for the specified band
 *
 * @param fd File descriptor of the open EDI file
 * @param band The band whose metadata is to be returned
 * @param metadata Pointer to the metadata_t structure that stores the data
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - EBAND: The band number passed was negative or greater than the maximum band
 * 			count for the EDI.
 *
 * Other than the above, the function also returns the error codes returned by
 * pread().
 */
static int get_meta_rec(int fd, band_t band, ediv0_metadata_t* metadata) {
	int len;
	size_t offset;


	/* Validate the band */
	if (!is_valid_band(band)) {
		errno = EBAND;
		return -1;
	}

	/* Read the metadata record */
	offset = OFF_META + sizeof(ediv0_metadata_t) * band;
	len = pread(fd, metadata, sizeof(ediv0_metadata_t), offset);
	if (len < sizeof(ediv0_metadata_t)) {
		perror("Cannot read metadata ");
		return -1;
	}

	return 0;
}



/**
 * @brief Writes the Metadata record for the specified band
 *
 * @param fd File descriptor of the open EDI file
 * @param band The band whose metadata is to be updated
 * @param metadata Pointer to the metadata_t structure that stores the data
 *
 * @return 0 on success; -1 on failure
 *
 * Error Codes:
 * - EBAND: The band number passed was negative or greater than the maximum band
 * 			count for the EDI.
 *
 * Other than the above, the function also returns the error codes returned by
 * pwrite().
 */
static int set_meta_rec(int fd, band_t band, ediv0_metadata_t* metadata) {
	int len;
	size_t offset;


	/* Validate the band */
	if (!is_valid_band(band)) {
		errno = EBAND;
		return -1;
	}

	/* Read the metadata record */
	offset = OFF_META + sizeof(ediv0_metadata_t) * band;
	len = pwrite(fd, metadata, sizeof(ediv0_metadata_t), offset);
	if (len < sizeof(ediv0_metadata_t)) 
		return -1;

	return 0;
}
