/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file edicreate.c
 *
 * @brief Creates EDI files based on user specified parameters
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#include <stdio.h>
#include <stdint.h>
#include <linux/limits.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <emudiskimg.h>
#include <debug.h>
#include <emulator.h>
#include <ediversion.h>



/* -- Macro Definitions -- */
#define ARGS_MIN 6
#define ARGS_MAX 11



/* -- Type Definitions -- */
typedef struct {
    char edi_path[PATH_MAX];
    char edi_metadata_path[PATH_MAX];
    uint32_t version;

    uint32_t blk_size;
    band_t band_count;
    rba_t band_size;
} cmd_args_t;



/* -- Constants -- */
const char* usage = "edicreate -b <band size> -c <band count> -s <block size> "
    "-v <EDI version> -f <EDI filename>";



/* -- Static Variables -- */
static cmd_args_t args;



/* -- Function Implementations -- */

static void print_usage() {
    /* Display the usage string */
    printf("\n%s\n\n", usage);
}



static void parse_args(int argc, char** argv) {
    int option, badargs = 0;


    /* Check if arguments were passed */
    if (argc < ARGS_MIN) {
        print_usage();
        exit(EXIT_FAILURE);
    }


    /* Loop over the arguments and parse them */
    while ((option = getopt(argc, argv, "b:c:s:f:v:m:")) != -1) {
        switch(option) {
        case 'b':                     /* Band Size */
            args.band_size = atoi(optarg);

            /* Ensure that the band size is greater than 0 */
            if (args.band_size <= 0) {
                printf("Band size should be greater than 0.\n");
                badargs = 1;
            }

            break;


        case 'c':                     /* Band Count */
            args.band_count = atoi(optarg);

            /* Ensure that the band count is greater than 0 */
            if (args.band_count <= 0) {
                printf("Band count should be greater than 0.\n");
                badargs = 1;
            }

            break;


        case 's':                     /* Block size */
            args.blk_size = atoi(optarg);

            /* Ensure that the block size is a power of 2 */
            if ((args.blk_size & (args.blk_size - 1)) != 0) {
                printf("Block size should be a power of 2.\n");
                badargs = 1;
            }

            break;


        case 'v':                     /* EDI Version */
            args.version = atoi(optarg);

            /* Validate the version */
            if (!ediv_is_valid(args.version)) {
                printf("Version %d is invalid", args.version);
                badargs = 1;
            }

            break;


        case 'f':                     /* EDI Filename */
            strcpy(args.edi_path, optarg);

            break;


	case 'm':		     /* EDI Metadata filename */
#ifdef _EDI_ON_RAW
	    strcpy(args.edi_metadata_path, optarg);
#endif
	    break;

        default:                      /* Unsupported option */
            printf("Invalid option: -%c", option);
            break;
        }
    }


    /* Ensure that all arguments were parsed successfully */
    if (badargs) {
        print_usage();
        exit(EXIT_FAILURE);
    }
}



int main(int argc, char** argv) {
    int fd;


    /* Parse the arguments passed */
    parse_args(argc, argv);
    dbg("Parsed arguments:\n\tMetadata File: %s\n\tFile: %s\n\tVersion: %d\n\tBlock Size: %d\n\t"
        "Band Count: %d\n\tBand Size: %d\n", args.edi_metadata_path, args.edi_path, args.version,
        args.blk_size, args.band_count,        args.band_size);


    /* Initialize the EDI Versioning module */
    if (ediv_init() < 0) {
        printf("Failed to initialize Verioning Module!\n");
        exit(EXIT_FAILURE);
    }


    /* Check the version number */
    switch(args.version) {
    case 0:            /* EDI v0 */
#ifdef _EDI_ON_RAW
        fd = edi_create(args.edi_metadata_path, args.edi_path, args.version,
                        args.blk_size, args.band_count,       args.band_size);
#else
	fd = edi_create(args.edi_path, args.version,
                        args.blk_size, args.band_count,       args.band_size);
#endif
        break;


    default:   /* Invalid version. Should not be here if the code is correct */
        printf("Unsupported EDI version %d.\n", args.version);
        print_usage();
        break;
    }


    /* Check if the file was successfully created */
    if (fd < 0) {
        printf("Error creating EDI file!\n");
	perror("edicreate");
        exit(EXIT_FAILURE);
    }
    dbg("Created v0 EDI file! Open with file descriptor %d\n", fd);


    /* Close the open EDI file */
    if (close(fd) < 0) {
        printf("EDI file created successfully, but unable to close it!\n");
        exit(EXIT_FAILURE);
    }

    printf("EDI file %s successfully created!\n", args.edi_path);
    return 0;
}
