/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file edidump.c
 *
 * @brief Dumps the metadata for a given EDI file
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <emudiskimg.h>
#include <debug.h>
#include <emulator.h>
#include <ediversion.h>



/* -- Constants -- */
#ifdef _EDI_ON_RAW
const char* usage = "edidump <EDI Metadata filename> <EDI filename>";
#else
const char* usage = "edidump <EDI filename>";
#endif



/* -- Function Implementations -- */

int main(int argc, char** argv) {
    int fd;
#ifdef _EDI_ON_RAW
    int meta_fd;
#endif

    /* Ensure that the EDI file path was passed as an argument */
    if (argc < 2) {
        fprintf(stderr, "%s\n\n", usage);
        exit(EXIT_FAILURE);
    }
    dbg("Parsed arguments:\n\tFile: %s\n", argv[1]);


    /* Initialize the EDI Versioning module */
    if (ediv_init() < 0) {
        printf("Failed to initialize Versioning Module!\n");
        exit(EXIT_FAILURE);
    }


    /* Open the EDI file */
#ifdef _EDI_ON_RAW
    if ((fd = edi_open(argv[1], argv[2], &meta_fd)) < 0) {
#else
    if ((fd = edi_open(argv[1])) < 0) {
#endif
        printf("Unable to open EDI file: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }


    /* Dump the EDI meta data */
#ifdef _EDI_ON_RAW
    if (edi_dump(meta_fd) < 0)
#else
    if (edi_dump(fd) < 0)
#endif
        printf("Unable to dump EDI file metadata: %s\n", strerror(errno));

    /* Close the open edi file */
#ifdef _EDI_ON_RAW
    edi_close(meta_fd, fd);
#else
    edi_close(fd);
#endif
    return EXIT_SUCCESS;
}

