/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define _XOPEN_SOURCE 500
#include <unistd.h>


#include "band_log.h"
#include "band_bitmap.h"
#include "cleaning_thread.h"
#include "common.h"
#include "band_meminfo.h"
#include "stats.h"

/*
 * @brief : Returns the bands to clean and band to clean to.
 *
 * @param : band_clean_t to store band number of source and destination of cleaning
 *          policy: which cleaning policy is chosen
 * @return : 0 : If some band is selected
 *          -1 : If no band is selected.
 *
 */
 
int select_cleaning_band(band_clean_t* band_clean, band_cleaning_policy_e policy)
{
    band_t band_to_clean;
    band_t band_clean_to;
    band_t band_with_max_deleted_so_far, band_with_min_effort_so_far;
    size_t free_size = 0, frag_size = 0, used_size = 0, clean_to_used = 0;
    int retval = -1;
    double ratio = 0;    
    time_t mod_time = 0;
    switch(policy) {
        case GET_CLEAN_BAND_FIRST:
            band_clean_to = band_meminfo_get_max_free_band(&free_size);

            if (band_clean_to == 0) {
                band_to_clean = band_meminfo_get_max_fragmented_band(&frag_size, &used_size);
                if (band_to_clean == 0) {
                    //no need to clean. no dead block
                    retval = -1;
                    break;
                }
            } else {
                band_to_clean = band_meminfo_get_fragmented_band(free_size, &frag_size, &used_size);
                if (band_to_clean == 0) {
                    band_meminfo_release_band_lock(band_clean_to);
                    if(frag_size == 0) {
                        //no need to clean. no dead block
                        retval = -1;
                        break;
                    } else {
                        //in place cleaning
                        band_to_clean = band_meminfo_get_max_fragmented_band(&frag_size, &used_size);
                        band_clean_to = 0;
                        retval = 0;
                    }
                }
            }

            band_clean->band_to_clean = band_to_clean;
            band_clean->band_clean_to = band_clean_to;
            retval = 0;
            break;

        case GET_BAND_TO_CLEAN_FIRST:
            band_to_clean = band_meminfo_get_max_fragmented_band(&frag_size, &used_size);
            if (band_to_clean == 0) 
                break;
            if (frag_size == 0) {
		band_meminfo_release_band_lock(band_to_clean);
                break;//no need to clean. No deleted fraction on band
            }
            band_clean_to = band_meminfo_get_clean_to_band(used_size, NULL,  &clean_to_used);
            band_clean->band_to_clean = band_to_clean;
            band_clean->band_clean_to = band_clean_to;
            retval = 0;
            break;

        case COST_BENEFIT:
            //cost benefit ratio is defined as
            //(1-utilization)*age/(1+utilization)
            //where age = timestamp of now - timestamp of last modified file on the band
            //utilization = used_size (size of live data)/ band_size
            band_to_clean = band_meminfo_get_max_cost_benefit_ratio_band(&ratio, &used_size, &frag_size, &mod_time);
            if (band_to_clean == 0)
                break;
	    if (frag_size == 0) {
	    	band_meminfo_release_band_lock(band_to_clean);
                break;//no need to clean. No deleted fraction on band
            }
            band_clean_to = band_meminfo_get_clean_to_band(used_size, NULL, NULL);
            band_clean->band_to_clean = band_to_clean;
            band_clean->band_clean_to = band_clean_to;
            retval = 0;
            break;

	case GREEDY_BY_MAX_SPACE_RECOVERED:
	    band_with_max_deleted_so_far = -1;
	    retval = 0;
	    while(1) {
		    band_to_clean = band_meminfo_get_max_deleted_band(band_with_max_deleted_so_far, &used_size);
		    if (band_to_clean == 0) {
			/*
			 * There are no bands having more deletions than band 0.
			 * This means that everything except 0 is fully free, i.e. nothing to clean.
			 */
			retval = -1;
                	break;
		    }
		    if (band_to_clean == -1) {
			/*
			 * No band's deleted capacity can fit into any other band.
			 * Cleaning cannot happen. Exit and panic.
			 */
			 retval = -1;
			 break;
		    }
		    band_clean_to = band_meminfo_get_clean_to_band(used_size, NULL, &clean_to_used);
		    if(band_clean_to != INVALID_BAND_NUM) {
			band_clean->band_to_clean = band_to_clean;
			band_clean->band_clean_to = band_clean_to;
			break;
		    }
		    else {
			band_with_max_deleted_so_far = band_to_clean;
			used_size = 0;
		    }
	    }
	    break;

	case GREEDY_BY_MIN_EFFORT:
	    band_with_min_effort_so_far = -1;
            retval = 0;
	    while(1) {
		band_to_clean = band_meminfo_get_min_effort_band(band_with_min_effort_so_far, &used_size);

		if (band_to_clean == 0) {
		    retval = -1;
		    break;
		}
		band_clean_to = band_meminfo_get_clean_to_band(used_size, NULL, &clean_to_used);
		if(band_clean_to != INVALID_BAND_NUM) {
		    band_clean->band_to_clean = band_to_clean;
		    band_clean->band_clean_to = band_clean_to;
		    break;
		}
		else {
		    band_with_min_effort_so_far = band_to_clean;
		    used_size = 0;
		}
	    }
	    break;
	
	case CLUSTERED_CLEANING_BY_PRESERVING_NAMESPACE:

        default:
            band_clean->band_clean_to = 0;
            band_clean->band_to_clean = 0;
            retval = -1;
            break;
    }

    return retval;
}    

int time_to_stop(int cleaned_count, int num_bands_to_clean, time_to_stop_policy_e policy) {
    int retval = 0;
    int fs_full;
    if ( num_bands_to_clean != 0 && cleaned_count > num_bands_to_clean) {
        log(LOG_CLN_INFO, "TIME_TO_STOP: reached band_to_clean limit %d out of %d\n", cleaned_count, num_bands_to_clean);
        return 1;
    }
    switch(policy) {
        case FIXED_THRESHOLD:
            stats_update();
            fs_full = stats_disk_utilization();
            log(LOG_CLN_INFO, "TIME_TO_STOP? %d\n", fs_full);
            if (fs_full <= FS_FULL_THRESHOLD_SOFT) {
    	        log(LOG_CLN_INFO, "TIME_TO_STOP: utilization %d <= SOFT %d\n", 
                          fs_full, FS_FULL_THRESHOLD_SOFT);

                retval = 1;
            }
            break;
        case MAX_FREE_SIZE:
            retval = 1; 
            break;
        default:
            break;
    }
    return retval;
}
