/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file band_meminfo.c
 *
 * @brief Implementation of in-memory data structures containing the
 *        band memory usage information
 *
 * @author Praveen Kumar Ramakrishnan <praveenr@andrew.cmu.edu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <float.h>
#include <string.h>
#include "failstop.h"
#include "band_meminfo.h"
#include "stats.h"
#include "helper.h"

#define UPDATE_AGE 1
#define STATE 1
//#define threshold_full SFSDATA.edi_bandsize * 0.1

static band_meminfo_t meminfo;

pthread_mutex_t meminfo_mutex;

int band_get_used_count(band_t band_num) {
    return meminfo.bands[band_num].used;
}

int band_get_free_count(band_t band_num) {
    return meminfo.bands[band_num].free;
}

int init_band_meminfo(void)
{
    int i;

    assert(SFSDATA.edi_bandcount > 0);

    if(pthread_mutex_init(&meminfo_mutex, NULL) != 0)
        return -1;

    meminfo.bands = (band_data_t *)malloc((SFSDATA.edi_bandcount) *
                                         sizeof(band_data_t));
    if(meminfo.bands == NULL)
        return -1;

    for(i = 0; i < SFSDATA.edi_bandcount; i++) {
        meminfo.bands[i].band_size = SFSDATA.edi_bandsize;
        meminfo.bands[i].free = SFSDATA.edi_bandsize;
        meminfo.bands[i].used = 0;
        meminfo.bands[i].deleted = 0;
        meminfo.bands[i].being_used = FALSE;
 
#ifdef STATE
        meminfo.bands[i].state_new = 0;
        meminfo.bands[i].state_clean_to = 0;
        meminfo.bands[i].state_delete = 0;
        meminfo.bands[i].state_cleaned = 0;
#endif
 
        meminfo.bands[i].last_mod_time = time(NULL);

        if(pthread_rwlock_init(&meminfo.bands[i].rwlock, NULL) != 0) {
            free(meminfo.bands);
            return -1;
        }
        meminfo.bands[i].lock_count = 0;
    }
    meminfo.disk_max_blocks = ((SFSDATA.edi_bandcount - 1) * SFSDATA.edi_bandsize);
    meminfo.clean_to_band = 1;
    stats_set(DISK_MAX_BLKS, meminfo.disk_max_blocks);
    stats_set(DISK_FREE_BLKS, meminfo.disk_max_blocks);
    stats_set(CLEANED_BLKS, 0);
    stats_set(CUM_CLEANED_BLKS, 0);
    stats_set(MOVED_BLKS, 0);
    stats_set(CUM_MOVED_BLKS, 0);

    return 0;
}


void cleanup_band_meminfo()
{
    int i;

    assert(meminfo.bands != NULL);

    for(i = 0; i < SFSDATA.edi_bandcount; i++) {
        assert(VALID_BAND_MEMINFO(&(meminfo.bands[i])));
        pthread_rwlock_destroy(&meminfo.bands[i].rwlock);
    }

    pthread_mutex_destroy(&meminfo_mutex);

    free(meminfo.bands);
}


void band_meminfo_increment_blocks_used(band_t band_num, size_t blocks)
{
    band_data_t *band = NULL;

    assert(band_num < SFSDATA.edi_bandcount);

    if(band_num == INVALID_BAND_NUM)
        return;

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));

    assert(blocks <= band->free);
    band->free -= blocks;

    band->used += blocks;

    stats_decr(DISK_FREE_BLKS, blocks);
                       
#ifdef STATE
    band->state_new= 1;
#endif

/* 
#ifdef UPDATE_AGE
    meminfo.bands[band_num].last_mod_time = time(NULL);
#endif
*/
}

#ifdef STATE
void band_meminfo_reset_state()
{  
    int i;
    for(i = 0; i < SFSDATA.edi_bandcount; i++) {
           meminfo.bands[i].state_new = 0;
           meminfo.bands[i].state_clean_to = 0;
           meminfo.bands[i].state_delete = 0;
           meminfo.bands[i].state_cleaned = 0;
    }
}
#endif

void band_meminfo_increment_blocks_deleted(band_t band_num, size_t blocks)
{
    band_data_t *band = NULL;

    assert(band_num < SFSDATA.edi_bandcount);

    if(band_num == INVALID_BAND_NUM)
        return;

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));

    assert(blocks <= band->used);
    band->used -= blocks;

    band->deleted += blocks;
    assert(VALID_BAND_MEMINFO(band));

#ifdef STATE
    band->state_delete = 1;
#endif

/*        
#ifdef UPDATE_AGE
    meminfo.bands[band_num].last_mod_time = time(NULL);
#endif
*/
}


void band_meminfo_update_band_cleaned(band_t band_num)
{
    band_data_t *band = NULL;
    size_t cleaned_blocks = 0;

    assert(band_num < SFSDATA.edi_bandcount);

    if(band_num == INVALID_BAND_NUM)
        return;

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));

    cleaned_blocks = band->used + band->deleted;

    stats_incr(CUM_CLEANED_BLKS, band->deleted);
    stats_set(CLEANED_BLKS, band->deleted);

    band->free = band->band_size;
    band->used = 0;
    band->deleted = 0;
 
#ifdef STATE
    band->state_cleaned = 1;
#endif
 
    stats_incr(DISK_FREE_BLKS, cleaned_blocks);
    stats_incr(CUM_MOVED_BLKS, cleaned_blocks);
    stats_set(MOVED_BLKS, cleaned_blocks);
    assert(stats_get(DISK_FREE_BLKS) <= meminfo.disk_max_blocks);
/*
#ifdef UPDATE_AGE
    meminfo.bands[band_num].last_mod_time = time(NULL);
#endif
*/
}


band_t band_meminfo_get_fragmented_band(size_t max_used_size, size_t *frag_size, size_t *used_size)
{
    size_t max = 0;
    size_t max_fragmented = 0;
    band_t i, band_num = INVALID_BAND_NUM;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);

    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used)
            continue;
        if(i == meminfo.clean_to_band || i == meminfo.last_written_band)
            continue;//Don't clean these two special band

        if(max_used_size != 0 &&
           meminfo.bands[i].used > max_used_size) {
            if(meminfo.bands[i].deleted > max_fragmented)
                max_fragmented = meminfo.bands[i].deleted;       
            continue;
        }

        if(meminfo.bands[i].deleted > max) {
            max = meminfo.bands[i].deleted;
            band_num = i;
        }
    }

    if(band_num != INVALID_BAND_NUM) {
        band_meminfo_lock_band_for_write(band_num);

        if(frag_size != NULL) {
            *frag_size = max;
        }
        if(used_size != NULL) {
            *used_size = meminfo.bands[band_num].used;
        }
    } else {
        if(frag_size != NULL) {
            *frag_size = max_fragmented;
        }
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    return band_num;
}

band_t band_meminfo_get_best_fit_band(size_t size_required, size_t *used_size)
{
    size_t min_waste = SFSDATA.edi_bandsize;
    size_t waste = 0;
    band_t i, band_num = INVALID_BAND_NUM;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);

    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used)
            continue;

        if(size_required != 0 &&
           meminfo.bands[i].free < size_required) {
            continue;
        }

        waste = meminfo.bands[i].free + meminfo.bands[i].deleted 
                -size_required;
        if(waste < min_waste) {
            min_waste = waste;       
            band_num = i;
        }
    }

    if(band_num != INVALID_BAND_NUM) {
        band_meminfo_lock_band_for_write(band_num);

        if(used_size != NULL) {
            *used_size = meminfo.bands[band_num].used;
        }
    } else {
        if(used_size != NULL) {
            *used_size = waste;
        }
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    return band_num;
}
 
band_t band_meminfo_get_age_sort_band(size_t size_required, time_t mod_time, size_t *used_size)
{
    double min_diff = DBL_MAX;
    double age_diff = 0;
    band_t i, band_num = INVALID_BAND_NUM;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);

    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used)
            continue;

        if(size_required != 0 &&
           meminfo.bands[i].free < size_required) {
            continue;
        }

        age_diff = difftime(meminfo.bands[i].last_mod_time, mod_time);
        if(age_diff < 0)
            age_diff = -age_diff;
        if(age_diff < min_diff) {
            min_diff = age_diff;       
            band_num = i;
        }
    }

    if(band_num != INVALID_BAND_NUM) {
        band_meminfo_lock_band_for_write(band_num);

        if(used_size != NULL) {
            *used_size = meminfo.bands[band_num].used;
        }
    } else {
        if(used_size != NULL) {
            *used_size = min_diff;
        }
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    return band_num;
}
 
band_t band_meminfo_get_max_fragmented_band(size_t *frag_size, size_t *used_size)
{
    return band_meminfo_get_fragmented_band(0, frag_size, used_size);
}

band_t band_meminfo_get_max_deleted_band(band_t band_with_max_deleted_so_far, size_t *used_size)
{
    band_t i, band_num = INVALID_BAND_NUM, band_num_max_used;
    size_t max_deleted = 0, max_used = 0;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);
    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
	if(meminfo.bands[i].being_used)
	    continue;
	if(i == meminfo.clean_to_band || i == meminfo.last_written_band)
	    continue;//Don't clean these two special band

	if((i != INVALID_BAND_NUM) && (meminfo.bands[i].deleted > max_deleted)) {
	    if(band_with_max_deleted_so_far != -1) {
		if(meminfo.bands[i].deleted >= meminfo.bands[band_with_max_deleted_so_far].deleted) {
		    /*
		     * We have checked bands of deletion length as much as
		     * band_with_max_deleted_so_far and no band is free enough to clean them.
		     */
		    band_num = -1;
		    continue;
		}
	    }
	    if(meminfo.bands[i].used > max_used) {
		band_num_max_used = i;
	    }
	    max_deleted = meminfo.bands[i].deleted;
	    band_num = i;
	}
    }
	
    if(used_size != NULL && band_num != INVALID_BAND_NUM && band_num != -1) {
	if(max_deleted == 0) {
	    //band_num = band_num_max_used;
	    band_num = -1;
	} else {
	    band_meminfo_lock_band_for_write(band_num);
	}
	*used_size = meminfo.bands[band_num].used;
    }
    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);
    return band_num;
}

double cost_benefit_ratio(band_t i, time_t now)
{
    double utilization = 0;
    double ratio = 0;
    double age = 1.0;

    if(i>=1 && i < SFSDATA.edi_bandcount) {
        utilization = (double) meminfo.bands[i].used / (double) meminfo.bands[i].band_size;
#ifdef UPDATE_AGE
        //age is the age of the youngest file
        //so it is the time from the last_modified_time to now
        //If band has a larger age, it is colder
	//add 0.1 to avoid now == last_mod_time
        age = difftime(now, meminfo.bands[i].last_mod_time) + 0.1;
#endif
        ratio = ((1.0 - utilization) / (1.0 + utilization)) * age; 
        if( meminfo.bands[i].deleted == 0)
            ratio = 0;//Don't clean band that has no dead data
        return ratio;
    } else {
        return 0;
    }
}
                   
band_t band_meminfo_get_cost_benefit_ratio_band(size_t max_used_size, double *ratio, size_t *used_size, size_t *deleted_size, time_t *mod_time)
{
    double max = 0;
    double cost_benefit = 0;
    time_t now = time(NULL);
    band_t i, band_num = INVALID_BAND_NUM;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);

    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used)
            continue;
        if(i == meminfo.clean_to_band || i == meminfo.last_written_band)
            continue;//Don't clean these two special band
        cost_benefit = cost_benefit_ratio(i, now);
        if(max_used_size != 0 &&
           meminfo.bands[i].used > max_used_size) {
            continue;
        }

        if(cost_benefit > max) {
            max = cost_benefit;
            band_num = i;
        }
    }

    if(band_num != INVALID_BAND_NUM) {
        band_meminfo_lock_band_for_write(band_num);

        if(ratio != NULL) {
            *ratio = max;
        }
        if(used_size != NULL) {
            *used_size = meminfo.bands[band_num].used;
        }
        if(deleted_size != NULL) {
            *deleted_size = meminfo.bands[band_num].deleted;
        }
        if(mod_time != NULL) {
            *mod_time = meminfo.bands[band_num].last_mod_time;
        }
    } else {
        if(ratio != NULL) {
            *ratio = max;
        }
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    return band_num;
}
 
band_t band_meminfo_get_max_cost_benefit_ratio_band(double *ratio, size_t *used_size, size_t *deleted_size, time_t *mod_time)
{
    return band_meminfo_get_cost_benefit_ratio_band(0, ratio, used_size, deleted_size, mod_time);
}

band_t band_meminfo_get_max_free_band(size_t *free_size)
{
    size_t max = 0;
    band_t i, band_num = INVALID_BAND_NUM;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);
    
    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used)
            continue;
	if(i == meminfo.clean_to_band || i == meminfo.last_written_band)
	    continue;
        if(meminfo.bands[i].free > max) {
            max = meminfo.bands[i].free;
            band_num = i;
        }
    }

    if(band_num != INVALID_BAND_NUM) {
        band_meminfo_lock_band_for_write(band_num);
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    if(free_size != NULL) {
        *free_size = max;
    }

    return band_num;
}


band_t band_meminfo_get_free_band_for_write(size_t size_required,
                                            rba_t *cursor, 
                                            size_t *avail_free_size)
{
    band_t band_num = INVALID_BAND_NUM;
    band_t lastband = meminfo.last_written_band;
    size_t free_size = 0;
        
    assert(lastband < SFSDATA.edi_bandcount);

    stats_set(LAST_ACTIVITY_TIME, time(NULL));

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);
    
    if((lastband > 0) && 
       (meminfo.bands[lastband].free >= size_required) &&
       (meminfo.bands[lastband].being_used == FALSE)) {
        band_num = lastband;
        free_size = meminfo.bands[band_num].free;

        band_meminfo_lock_band_for_write(band_num);
    }
    else {
        assert(pthread_mutex_unlock(&meminfo_mutex) == 0);
        band_num = band_meminfo_get_max_free_band(&free_size);
        if(band_num == 0) {
            return INVALID_BAND_NUM;
        }
        if(free_size < size_required) {
            band_meminfo_release_band_lock(band_num);
            return INVALID_BAND_NUM;
        }
    }

    assert(band_num != INVALID_BAND_NUM);

    meminfo.last_written_band = band_num;
    
    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    if(cursor != NULL) {
        *cursor = meminfo.bands[band_num].used 
                + meminfo.bands[band_num].deleted;
    }

    if(avail_free_size != NULL) {
        *avail_free_size = free_size;
    }

    return band_num;
}

band_t band_meminfo_get_clean_to_band(size_t size_required,
                                            rba_t *cursor, 
                                            size_t *avail_free_size)
{
    band_t band_num = INVALID_BAND_NUM;
    band_t clean_to = meminfo.clean_to_band;
    size_t free_size = 0;
        
    assert(clean_to >= 0);
    assert(clean_to < SFSDATA.edi_bandcount);

    stats_set(LAST_ACTIVITY_TIME, time(NULL));

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);
    
    if((clean_to >= 0) && 
       (meminfo.bands[clean_to].free >= size_required) &&
       (meminfo.bands[clean_to].being_used == FALSE)) {
	
        band_num = clean_to;

        band_meminfo_lock_band_for_write(band_num);
        free_size = meminfo.bands[band_num].free;
#ifdef STATE
        meminfo.bands[band_num].state_clean_to = 1;
#endif
    }
    else {
        assert(pthread_mutex_unlock(&meminfo_mutex) == 0);
        band_num = band_meminfo_get_max_free_band(&free_size);
        if(band_num == 0) {
            meminfo.clean_to_band = band_num;
            //in place cleaning
            return INVALID_BAND_NUM;
        }
        if(free_size < size_required) {
            //not enough space for cleaning
            //should add return failure
            band_meminfo_release_band_lock(band_num);
            return INVALID_BAND_NUM;
        }

    	meminfo.clean_to_band = band_num;
#ifdef STATE
    	meminfo.bands[band_num].state_clean_to = 1;
#endif   
    }
    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);

    if(cursor != NULL) {
        *cursor = meminfo.bands[band_num].used 
                + meminfo.bands[band_num].deleted;
    }

    if(avail_free_size != NULL) {
        *avail_free_size = free_size;
    }

    return band_num;
}

band_t band_meminfo_get_min_effort_band(band_t band_with_min_effort_so_far,
					size_t *used_size) {
    int i = 0;
    band_t band_num = 0;
    size_t temp_used_size = 0, frag_size = 0;
    double effort = 0.0, max_effort = 0.0;

    assert(pthread_mutex_lock(&meminfo_mutex) == 0);
    
    for(i = 1; i < SFSDATA.edi_bandcount; i++) {
        if(meminfo.bands[i].being_used || i == meminfo.clean_to_band || i == meminfo.last_written_band || (band_with_min_effort_so_far != -1 && i == band_with_min_effort_so_far) || (meminfo.bands[i].used == 0 && meminfo.bands[i].deleted == 0))
            continue;

	if(meminfo.bands[i].deleted > 0 && meminfo.bands[i].used == 0) {
	    effort = (double)(meminfo.bands[i].deleted / 1.0);
	}
	else {
	    effort = (double)(meminfo.bands[i].deleted / meminfo.bands[i].used);
	}
        if(effort > max_effort) {
	    band_num = i;
	    max_effort = effort;
	}
    }
    
    if(band_num != INVALID_BAND_NUM) {
	if(used_size != NULL) {
	    *used_size = meminfo.bands[band_num].used;
	}
        band_meminfo_lock_band_for_write(band_num);
    }

    assert(pthread_mutex_unlock(&meminfo_mutex) == 0);
 
    return band_num;
}

int band_meminfo_lock_band_for_read(band_t band_num)
{
    band_data_t *band = NULL;

    assert(band_num < SFSDATA.edi_bandcount);

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));

    dbg("request lock band %d for read %d\n", band_num, band->lock_count);
    stats_set(LAST_ACTIVITY_TIME, time(NULL));
    assert(pthread_rwlock_rdlock(&band->rwlock) == 0);
    band->lock_count ++;
    dbg("success: lock band %d for read %d\n", band_num, band->lock_count);
    meminfo.bands[band_num].being_used = TRUE;

    return 0;
}

int band_meminfo_lock_band_for_write(band_t band_num)
{
    band_data_t *band = NULL;

    assert(band_num < SFSDATA.edi_bandcount);

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));

    stats_set(LAST_ACTIVITY_TIME, time(NULL));
    dbg("request lock band %d for write, count %d\n", band_num,
        band->lock_count);
    assert(pthread_rwlock_wrlock(&band->rwlock) == 0);
    meminfo.bands[band_num].being_used = TRUE;
    band->lock_count ++;
    dbg("success: lock band %d for write, count %d\n", band_num,
        band->lock_count);

    return 0;
}

int band_meminfo_release_band_lock(band_t band_num)
{
    band_data_t *band = NULL;

    assert(band_num < SFSDATA.edi_bandcount);

    band = &(meminfo.bands[band_num]);
    assert(VALID_BAND_MEMINFO(band));
    band->lock_count --;
    dbg("unlock band %d, count %d\n", band_num, band->lock_count);
    meminfo.bands[band_num].being_used = FALSE;
    assert(pthread_rwlock_unlock(&band->rwlock) == 0);
    return 0;
}

int band_meminfo_get_cleaner_stats(band_data_t * band)
{
    int i;
    for(i = 0; i < SFSDATA.edi_bandcount; i++) {
        assert(VALID_BAND_MEMINFO(&meminfo.bands[i]));
    } 
    if(band == NULL)
        return -1;
    else
        memcpy(band , meminfo.bands, SFSDATA.edi_bandcount*sizeof(band_data_t));
    return 0;
}
