/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file band_meminfo.h
 *
 * @brief Macros, data structures and function prototypes pertaining to the
 * in-memory band information.
 *
 * @author Praveen Kumar Ramakrishnan <praveenr@andrew.cmu.edu>
 */

#ifndef __BAND_MEMINFO_H_
#define __BAND_MEMINFO_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <fuse.h>
#include <pthread.h>
#include <time.h>

#include "common.h"

#define INVALID_BAND_NUM 0

typedef struct {
    size_t band_size;
    size_t used;
    size_t deleted;
    size_t free;
    int being_used;
    int state_new;
    int state_clean_to;
    int state_delete;
    int state_cleaned;
    pthread_rwlock_t rwlock;
    int lock_count;
    time_t last_mod_time;
} band_data_t;

typedef struct {
    band_data_t *bands;
    band_t last_written_band;
    band_t clean_to_band;
    size_t disk_max_blocks;
} band_meminfo_t;


#define VALID_BAND_MEMINFO(ptr) \
    ((ptr) != NULL && ((ptr)->band_size == \
     ((ptr)->used + (ptr)->deleted + (ptr)->free)))


int init_band_meminfo(void);
void cleanup_band_meminfo();

void band_meminfo_increment_blocks_used(band_t band_num, size_t blocks);
void band_meminfo_increment_blocks_deleted(band_t band_num, size_t blocks);
void band_meminfo_update_band_cleaned(band_t band_num);
void band_meminfo_reset_state();

band_t band_meminfo_get_max_fragmented_band(size_t *frag_size, size_t *used_size);
band_t band_meminfo_get_fragmented_band(size_t max_used_size, size_t *frag_size, 
                                        size_t *used_size);
band_t band_meminfo_get_max_cost_benefit_ratio_band(double *ratio, 
                        size_t *used_size, size_t *deleted, time_t *mod_time);
band_t band_meminfo_get_max_free_band(size_t *free_size);
band_t band_meminfo_get_free_band_for_write(size_t size_required,
                                            rba_t *cursor, 
                                            size_t *avail_free_size);
band_t band_meminfo_get_clean_to_band(size_t size_required,
                                            rba_t *cursor, 
                                            size_t *avail_free_size);
band_t band_meminfo_get_best_fit_band(size_t size_required, 
                                            size_t *avail_free_size);
band_t band_meminfo_get_age_sort_band(size_t size_required, 
                                time_t mod_time, size_t *used_size);

band_t band_meminfo_get_max_deleted_band(band_t band_with_max_deleted_so_far, 
					    size_t *used_size);
band_t band_meminfo_get_min_effort_band(band_t band_with_min_effort_so_far, 
					    size_t *used_size);

int band_meminfo_lock_band_for_read(band_t band_num);
int band_meminfo_lock_band_for_write(band_t band_num);
int band_meminfo_release_band_lock(band_t band_num);

int band_meminfo_get_cleaner_stats(band_data_t * band);

int band_get_used_count(band_t band_num);
int band_get_free_count(band_t band_num);
#ifdef __cplusplus
}
#endif
#endif
