/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file buffercache.c
 *
 * @brief Implements the buffer cache abstraction for ShingledFS
 *
 * Always grab the BC Files list lock first, before grabbing the lock on the
 * BC File record. This ordering results in deadlock prevention.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * @bug Hard coded the number of files that can be held in the buffer cache.
 * This value should be calculated as buffer_cache_size / EDI_band_size.
 */

#define _XOPEN_SOURCE 500
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/xattr.h>

#include "failstop.h"
#include "buffercache.h"
#include "common.h"
#include "helper.h"

#include "inode.h"
#include "band_bitmap.h"
#include "band_log.h"

#include "emulator.h"
#include "emutypes.h"
#include "emuerror.h"
#include "band_meminfo.h"
#include "cleaning_thread.h"


/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

/**
 * @brief Defines the maximum number of files in the buffer cache
 */
#define MAX_FILES                       256



/* ====================== */
/* -- Type Definitions -- */
/* ====================== */

/**
 * @brief Defines the Buffer Cache File List
 *
 * This structure abstracts away all of the details pertaining to the buffer
 * cache implementation.
 *
 * @param max_files Max. no. of files that can be present in the buffer cache
 * @param cached_files No. of files currently in the buffer cache
 * @param bc_files List of bc_file_t records for each file in the buffer cache
 * @param lock Lock that protects the structure during concurrent access
 */
typedef struct {
    int		max_files;
    int         cached_files;
    bc_file_t   *list;

    pthread_mutex_t  lock;
}bc_file_list_t;



/**
 * @brief Defines the request list that holds loads/stores from ShingledFS
 *
 * @param list List of requests to load/store a file
 * @param lock Lock that protects the structure during concurrent access
 */
typedef struct {
    bc_request_t	*list;
    sem_t               count;
    pthread_mutex_t  	lock;
}bc_request_list_t;



/* ====================== */
/* -- Static Variables -- */
/* ====================== */
static bc_file_list_t bc_files;
static bc_request_list_t bc_requests;



/* ========================= */
/* -- Function Prototypes -- */
/* ========================= */

static bc_file_t        *bc_load_file(const char *path, inode_t *inode);
static void             bc_prepare_tmpFS_file(const char *path, int *fd,
                                              int *fd_xattr, int inode_num);
static void          bc_evict_file(bc_file_t *file);

static bc_file_t        *bc_file_new(const char *path, int inode);
static void                      bc_file_enqueue(bc_file_t *file, int insert_at_head);
static void                      bc_file_dequeue(bc_file_t *file);
static bc_file_t    *bc_file_search_toevict();

static bc_request_t     *bc_enqueue_request(bc_file_t *file, bc_operation_t op);
static bc_request_t     *bc_dequeue_request();

static void                     *bc_worker();



/* ============================== */
/* -- Function Implementations -- */
/* ============================== */

/**
 * @brief Initializes the Buffer Cache system
 *
 * This function initializes the necessary data structures and spawns the worker
 * threads that handle data transfer between the buffer cache and the EDI.
 *
 * @param cache_size Size of the Buffer Cache in bytes
 */
void
bc_init(size_t cache_size) {
    pthread_t tid;


    /* Log the call */
    log(LOG_BC_CALLS, "bc_init(cache_size=%ld)\n", cache_size);


    /* Validate the inputs */
    if (cache_size == 0) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }

    /* Initialize the BC File list */
    bc_files.max_files = MAX_FILES;
    bc_files.cached_files = 0;
    bc_files.list = NULL;
    dod(pthread_mutex_init(&bc_files.lock, NULL) == 0);


    /* Initialize the BC Request list */
    bc_requests.list = NULL;
    dod(sem_init(&bc_requests.count, 0, 0) == 0);
    dod(pthread_mutex_init(&bc_requests.lock, NULL) == 0);


    /* Spawn the worker thread */
    dod(pthread_create(&tid, NULL, bc_worker, NULL) == 0);
    dod(pthread_detach(tid) == 0);
    log(LOG_SFS_ERRORS, "Buffer Cache set to a max. of %d files\n", MAX_FILES);
}

/*
 * This is similar to bc_clean - just that it evicts all the files from the
 * buffer cache
 * TODO: Find out why bc_clean is needed in the first place!!!
 */
int bc_destory()
{
    /*
     * Evict all files from the buffer cache,
     * if a file is in use, return in use error
     */
    bc_file_t *cur_file;
    /* Lock the BC Files list */
    assert(!pthread_mutex_lock(&bc_files.lock));

    /* Iterate over the BC File list to find the record for the file */
    while((cur_file = bc_files.list) != NULL) {
        assert(bc_files.cached_files > 0);

        /* Lock the current file record */
        dod(!pthread_mutex_lock(&(cur_file->mutex)));
        /* Check the file's ref-count */
        if (cur_file->ref_count == 0) {
            /* Check if the file has been stored */
            if (cur_file->state == ST_STORED) {
                log(LOG_BC_INFO, "Evicting %s! State: %d, "
                    "Ref_count: %d\n", cur_file->path,
                    cur_file->state, cur_file->ref_count);
                dod(!pthread_mutex_unlock(&(cur_file->mutex)));
            } else if ((cur_file->state == ST_LOADED) ||
                       (cur_file->state == ST_STORING)) {
                /* Ensure that the file is LOADED or STORING. It could be
                 * possible that the file has a request for STORING in the queue
                 * that has not been processed yet. So the state could still be
                 * LOADED
                 *
                 * @todo Verify the above. State is being changed from LOADED to
                 * STORING synchronously in shingledfs_close(). So this might be
                 * an incorrect assumption!
                 */

                /* Increment the ref_count to force the worker to signal when
                 * the file has been stored. */
                cur_file->ref_count++;

                /* Block waiting for the worker to finish writing the file */
                dod(!pthread_cond_wait(&cur_file->condvar, &cur_file->mutex));

                /* Close the file */
                close(cur_file->fd);
                close(cur_file->fd_xattr);

                /* Decrement the ref_count to 0 */
                cur_file->ref_count = 0;
                cur_file->fd = -1;
                cur_file->fd_xattr = -1;
                cur_file->state = ST_STORED;

                /* Unlock the current file record */
                dod(!pthread_mutex_unlock(&(cur_file->mutex)));
            } else {
                /* TODO: check if the else is needed. assert added to catch the case */
                assert(0);
                dod(!pthread_mutex_unlock(&(cur_file->mutex)));
            }

            /* Evict the file from the buffer cache */
            /* TODO: No locking in evict or de-queue - BUG: ? */
            bc_evict_file(cur_file);
        } else {
            /* Unlock the current file record */
            dod(!pthread_mutex_unlock(&(cur_file->mutex)));
            assert(!pthread_mutex_unlock(&bc_files.lock));
            /* Files in use, cannot un mount yet */
            errno = EBUSY;
            return -1;
        }
    }

    /* Unlock the BC Files list */
    assert(!pthread_mutex_unlock(&bc_files.lock));
    return 0;
}


/**
 * @brief Loads a file into the buffer cache
 *
 * @param path Path to the file to be loaded into the buffer cache
 *
 * @return Pointer to the BC File record, on success
 */
bc_file_t
*bc_open(const char *path) {
    int inode_no, fd, fd_xattr;
    inode_t inode;
    bc_file_t *cur_file, *retval;
    char path_xattr[PATH_MAX];


    /* Log the call */
    log(LOG_BC_CALLS, "bc_open(%s)\n", path);


    /* Validate the inputs */
    if (path == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Identify the inode for the file */
    dod(!xattr_get(path, X_INODE, &inode_no, sizeof(int)));
    dod(!get_inode(inode_no, &inode));
    log(LOG_BC_INFO, "Inode %d for %s: Band: %d, RBA: %d, Extent: %d\n",
        inode.inode_num,        path, inode.band_id, inode.band_offset,
        inode.size);


    /* Lock the BC Files list */
    dod(!pthread_mutex_lock(&bc_files.lock));


    /* Find the BC File record for the file */
    cur_file = bc_file_search_path(path);

    /* Check if the record was found */
    if (cur_file == NULL) {
        log(LOG_BC_INFO, "Not one of the %d files in the cache\n",
            bc_files.cached_files);

        /* Check if there is room to add more files in the cache */
        if (bc_files.max_files == bc_files.cached_files) {
            /* Check if there are any files that can be evicted. */
            dod((cur_file = bc_file_search_toevict()) != NULL);

            /* Evict the file */
            log(LOG_BC_INFO, "Buffer Cache full! Evicting %s\n",
                cur_file->path);
            bc_evict_file(cur_file);
        }

        /* Create the tmpFS file and its symlink on the Unshingled partition */
        bc_prepare_tmpFS_file(path, &fd, &fd_xattr, inode.inode_num);

        /* Insert the record into the list */
        dod((cur_file = bc_load_file(path, &inode)) != NULL);
        cur_file->fd = fd;
        cur_file->fd_xattr = fd_xattr;

        /* Signal to a worker to pick up the newly queued request */
        if (inode.band_id)
            dod(!sem_post(&(bc_requests.count)));
    }


    /* Update the ref-count */
    dod(!pthread_mutex_lock(&(cur_file->mutex)));
    cur_file->ref_count++;
    dod(!pthread_mutex_unlock(&bc_files.lock));
    log(LOG_BC_INFO, "%s created and opened in Buffer Cache. Ref-count: %d\n",
        cur_file->path, cur_file->ref_count);


    /* Check if the file needs to be loaded from the EDI */
    if ((inode.band_id != 0) || ((inode.band_id == 0) && (cur_file->dirty))) {
        /* Check if the file is being loaded or stored */
        while ((cur_file->state == ST_LOADING) ||
               (cur_file->state == ST_STORING)) {
            log(LOG_BC_INFO, "File state: %s... going to sleep!\n",
                (cur_file->state == ST_LOADING) ? "Loading" : "Storing");
            dod(!pthread_cond_wait(&(cur_file->condvar), &(cur_file->mutex)));
            log(LOG_BC_INFO, "Awoken by worker...\n");
        }
        dod((cur_file->state == ST_LOADED) || (cur_file->state == ST_STORED));
    }

    /* Check if the file has been stored. This condition is possible when a file
     * is created, opened, and closed without writing anything! */
    if (cur_file->state == ST_STORED) {
        /* Re-open the file. If there is an error, it will be handled! */
        log(LOG_BC_INFO, "File already loaded in Buffer Cache!\n");
        dod(cur_file->fd < 0);

        xattr_path(cur_file->path, path_xattr);
        dod((cur_file->fd = open(cur_file->path, O_RDWR)) > 0);
        dod((cur_file->fd_xattr = open(path_xattr, O_RDWR)) > 0);

        /* Update the file state */
        cur_file->state = ST_LOADED;

        /* Reposition the BC File record at the head of the BC File List */
        dod(!pthread_mutex_lock(&bc_files.lock));
        bc_file_dequeue(cur_file);
        bc_file_enqueue(cur_file, 1);
        dod(!pthread_mutex_unlock(&bc_files.lock));
    }

    /* Ensure that the file was loaded */
    retval = cur_file;
    dod(cur_file->fd > 0);
    dod(cur_file->fd_xattr > 0);

    /* Unlock the BC File record */
    dod(!pthread_mutex_unlock(&(cur_file->mutex)));

    return retval;
}



/**
 * @brief Saves a file in the buffer cache to disk
 *
 * @param file Pointer to the BC File record of the file that needs to be closed
 */
void
bc_close(bc_file_t *file) {
    inode_t inode;
    struct stat statbuf;


    /* Log the call */
    log(LOG_BC_CALLS, "bc_close(%s)\n", file->path);


    /* Validate the inputs */
    if (file == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Lock the BC Files list */
    dod(!pthread_mutex_lock(&bc_files.lock));

    /* Lock the BC File record */
    dod(!pthread_mutex_lock(&(file->mutex)));


    /* Decrement the ref-count */
    file->ref_count--;

    /* Check to see if this is the last reference to the file */
    if (file->ref_count == 0) {
        /* Check if the file is dirty. If so, hand it off to the worker. If not,
         * then handle the closing here. */
        if (file->dirty) {
            /* Mark the file state as STORING */
            file->state = ST_STORING;

            /* Reposition the file entry */
            bc_file_dequeue(file);
            bc_file_enqueue(file, 0);

            /* Enqueue a Request to write this file back to the EDI */
            log(LOG_BC_INFO, "Ref-count is 0; signaling worker to save file\n");
            dod(bc_enqueue_request(file, OP_STORE) != NULL);

            /* Signal to a worker to pick up the newly queued request */
            dod(!sem_post(&(bc_requests.count)));

        } else {
            /* Check if the file was deleted */
            if (file->deleted) {

                /* Read the file inode */
                dod(!get_inode(file->inode, &inode));
                log(LOG_BC_INFO, "File has been deleted while open. Freeing "
                    "file inode %d: Band: %d, RBA: %d, Extent: %d"
                    ", Inode No.: %d, Append Log Entry No.: %d\n",
                    file->inode, inode.band_id,
                    inode.band_offset, inode.size, inode.inode_num,
                    inode.entrynumber);

                /* Unlink the inode */
                while( inode_unlink(file->inode, inode) != 0)
                    dod(!get_inode(file->inode, &inode));

		inode_xattr_free(file->inode);
            }

            /* Mark the file as clean */
            file->dirty = 0;

            /* Copy the file's stats to disk */
            if (!file->deleted) {
                dod(!fstat(file->fd, &statbuf));
                dod(!fsetxattr(file->fd_xattr, X_STATS, &statbuf,
                               sizeof(struct stat), 0));
            }

            /* Mark the state as STORED and close the open files */
            close(file->fd);
            close(file->fd_xattr);
            file->fd = -1;
            file->fd_xattr = -1;
            log(LOG_BC_INFO, "Closed %s on tmpFS.\n", file->path);
            file->state = ST_STORED;
        }
    }


    /* Unlock the BC File record */
    dod(!pthread_mutex_unlock(&(file->mutex)));

    /* Unlock the BC File list */
    dod(!pthread_mutex_unlock(&bc_files.lock));
}



/**
 * @brief Deletes the specified file from the buffer cache.
 *
 * @param path Path to the file to be unloaded
 *
 * @return 0 on success; -1 on failure due to transitive symlinks
 */
int
bc_unlink(const char *path) {
    bc_file_t *cur_file;
    char path_bc[PATH_MAX];
    int link_len;

    /* Log the call */
    log(LOG_BC_CALLS, "bc_unlink(path=%s)\n", path);


    /* Validate the inputs */
    if (path == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Generate the tmpFS file path for the path */
    link_len = readlink(path, path_bc, PATH_MAX);
    path_bc[link_len] = '\0';


    /* Lock the BC File List */
    dod(!pthread_mutex_lock(&bc_files.lock));

    /* Search for the BC File record for the file */
    if ((cur_file = bc_file_search_path(path)) == NULL) {
        log(LOG_BC_INFO, "%s not present in BC!\n", path_bc);
        dod(!pthread_mutex_unlock(&bc_files.lock));
        return -1;
    }

    /* Lock the file record */
    dod(!pthread_mutex_lock(&(cur_file->mutex)));

    /* Check the state of the file */
    if (cur_file->state == ST_STORED) {
        /* Dequeue the BC File record */
        bc_file_dequeue(cur_file);
        log(LOG_BC_INFO, "Deleted %s from Buffer Cache\n", cur_file->path);
        free(cur_file);
        goto cleanup_file;

    } else {
        /* Mark the file as deleted */
        cur_file->deleted = 1;
        strcat(cur_file->path, "(deleted)");
        log(LOG_BC_INFO, "Set deleted to TRUE in BC File record for %s\n",
            cur_file->path);

    }

    dod(!pthread_mutex_unlock(&(cur_file->mutex)));

cleanup_file:
    /* Delete the tmpFS copy of the file */
    dod(!unlink(path_bc));
    log(LOG_BC_INFO, "Deleted %s from tmpFS!\n", path_bc);

    /* Unlock the BC File List */
    dod(!pthread_mutex_unlock(&bc_files.lock));

    return 0;
}



/**
 * @brief Renames the specified file in the buffer cache
 *
 * @param path Path to the existing file
 * @param path_new Path to the renamed file
 *
 * @return 0 on success; -1 on failure due to transitive symlinks
 */
int
bc_rename(const char    *path,
          const char    *path_new) {
    char path_bc[PATH_MAX], path_bc_new[PATH_MAX];
    bc_file_t *cur_file, *renamed_file;
    inode_t inode;
    int link_len;


    /* Log the call */
    log(LOG_BC_CALLS, "bc_rename(path=%s, path_new=%s)\n", path, path_new);


    /* Validate the inputs */
    if ((path == NULL) || (path_new == NULL)) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Generate the tmpFS file path for the path */
    link_len = readlink(path, path_bc, PATH_MAX);
    path_bc[link_len] = '\0';


    /* Lock the BC File List */
    dod(!pthread_mutex_lock(&bc_files.lock));

    /* Search for the BC File record for the file */
    if ((cur_file = bc_file_search_path(path)) == NULL) {
        log(LOG_BC_INFO, "%s not present in BC!\n", path_bc);
        assert(!pthread_mutex_unlock(&bc_files.lock));
        return -1;
    }

    /* Store a pointer to the BC File Record of the file to be renamed */
    renamed_file = cur_file;
    log(LOG_BC_INFO, "File to rename: Path: %s, Inode No: %d, Descriptor: %d, "
        "RefCount: %d, State: %d, Dirty: %d, Deleted: %d\n",
        renamed_file->path, renamed_file->inode, renamed_file->fd,
        renamed_file->ref_count, renamed_file->state,
        renamed_file->dirty, renamed_file->deleted);


    /* Ensure that the renamed file does not exist in the Buffer Cache. If so,
     * its entry needs to be removed because the rename() function will
     * atomically remove it. */
    if ((cur_file = bc_file_search_path(path_new)) != NULL) {
        log(LOG_BC_INFO, "File target name: Path: %s,  Inode No: %d, "
            "Descriptor: %d, RefCount: %d, State: %d, Dirty: %d, "
            "Deleted: %d\n", cur_file->path, cur_file->inode,
            cur_file->fd, cur_file->ref_count, cur_file->state,
            cur_file->dirty, cur_file->deleted);

        /* Check the state of the file */
        dod(!pthread_mutex_lock(&(cur_file->mutex)));
        if (cur_file->state == ST_STORED) {
            /* Read the inode information for the file */
            dod(!get_inode(cur_file->inode, &inode));

            /* Unlink the inode */
            while( inode_unlink(cur_file->inode, inode)!= 0)
                dod(!get_inode(cur_file->inode, &inode));

	    inode_xattr_free(cur_file->inode);

            log(LOG_BC_INFO, "Freed inode %d\n", cur_file->inode);

            /* Dequeue the BC File record */
            bc_file_dequeue(cur_file);
            log(LOG_BC_INFO, "Deleted BC File record for %s\n", cur_file->path);
            free(cur_file);
            goto cleanup_file;

        } else {
            /* Mark the file as deleted */
            cur_file->deleted = 1;
            strcat(cur_file->path, "(deleted)");
            log(LOG_BC_INFO, "Set deleted to TRUE in BC File record for %s\n",
                cur_file->path);

        }
        dod(!pthread_mutex_unlock(&(cur_file->mutex)));
    }


cleanup_file:
    /* Update the name of the file in its BC File record */
    dod(!pthread_mutex_lock(&(renamed_file->mutex)));
    strcpy(renamed_file->path, path_new);
    dod(!pthread_mutex_unlock(&(renamed_file->mutex)));
    log(LOG_BC_INFO, "Updated BC Record for original file to new name %s\n",
        renamed_file->path);


    /* Unlock the BC File List */
    dod(!pthread_mutex_unlock(&bc_files.lock));


    /* Generate the tmpFS file path for the path */
    bc_generate_tmpFS_path(path_bc_new, path_new, renamed_file->inode);

    /* Rename the tmpFS copy of the file */
    dod(!rename(path_bc, path_bc_new));
    log(LOG_BC_INFO, "%s renamed to %s\n", path_bc, path_bc_new);


    return 0;
}



/**
 * @brief Cleans the Buffer Cache by removing a file that can be evicted.
 *
 * This function is called when the Buffer Cache has no more space to write data
 * to. Essentially, it searches for a file that can be evicted, and removes it
 * from the buffer cache. The space occupied by the file is consequently freed
 * up for use.
 */
void
bc_clean() {
    bc_file_t *cur_file;


    /* Lock the BC Files list */
    assert(!pthread_mutex_lock(&bc_files.lock));

    /* Check if there are any files that can be evicted using an LRU
     * heuristic. */
    log(LOG_BC_INFO, "Buffer Cache full. Searching for a file to evict.\n");
    dod((cur_file = bc_file_search_toevict()) != NULL);

    /* Evict the file */
    log(LOG_BC_INFO, "Evicting %s\n", cur_file->path);
    bc_evict_file(cur_file);

    /* Unlock the BC Files list */
    dod(!pthread_mutex_unlock(&bc_files.lock));
}



/* ===================================== */
/* -- Static Function Implementations -- */
/* ===================================== */

/**
 * @brief Creates and adds a new BC File record to the BC Files list.
 *
 * @param path Path to the file to be loaded into the BC
 * @param inode Inode number of the file to be loaded into the BC
 *
 * @return Pointer to the newly created BC File record
 */
static bc_file_t
*bc_load_file(const char        *path,
              inode_t           *inode) {
    bc_file_t *new_file;


    /* Validate the inputs */
    if ((path == NULL) || (inode == NULL)) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Create a new cached file record */
    new_file = bc_file_new(path, inode->inode_num);

    /* Enqueue the new file in the BC Files list */
    bc_file_enqueue(new_file, 1);

    /* Enqueue a File Load request to load the new file from the EDI */
    if (inode->band_id != 0) {
        bc_enqueue_request(new_file, OP_LOAD);
        log(LOG_BC_INFO, "Enqueued request to worker to load file from EDI.\n");
    } else
        log(LOG_BC_INFO, "Newly created file. No need to call the worker.\n");

    return new_file;
}



/**
 * @brief Prepares the file for loading into tmpFS.
 *
 * This function creates a new file on tmpFS and replaces the regular file on
 * the unshingled partition with a symbolic link to the newly created file on
 * tmpFS. This allows for transparent transmission of file operations from the
 * unshingled partition to the tmpFS file.
 *
 *
 * @param path Path to the file on the unshingled partition
 * @param fd File descriptor to the open file on tmpFS
 * @param fd_xattr File descriptor to the Attribute Replication file
 */
static void
bc_prepare_tmpFS_file(const char *path,
                      int *fd, int *fd_xattr, int inode_num) {

    char path_bc[PATH_MAX];
    struct stat statbuf;


    /* Validate the inputs */
    if ((path == NULL) || (fd == NULL) || (fd_xattr == NULL)) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Open the Attribute Replication file */
    xattr_path(path, path_bc);
    dod((*fd_xattr = open(path_bc, O_RDWR)) > 0);

    /* Read the file stats */
    dod(fgetxattr(*fd_xattr, X_STATS, &statbuf, sizeof(struct stat)) >= 0);

    /* Generate the tmpFS file path */
    bc_generate_tmpFS_path(path_bc, path, inode_num);
    log(LOG_BC_INFO, "Generated tmpFS filename %s\n", path_bc);


    /* Create the file in tmpFS */
    dod((*fd = open(path_bc, O_CREAT|O_RDWR, statbuf.st_mode)) > 0);

    /* Delete the regular file */
    dod(!unlink(path));

    /* Create a symbolic link to the tmpFS file */
    dod(!symlink(path_bc, path));
    log(LOG_BC_INFO, "Symlink %s created; pointing to %s on descriptor %d\n",
        path, path_bc, *fd);
}



/**
 * @brief Evicts a file from the Buffer Cache
 *
 * This function expects that the file to be evicted has been checked to ensure
 * that there are no open file descriptors on the file. It replaces the symlink
 * on the Unshingled partition with a regular file, updates the attributes of
 * the file in the Attribute Replication file, updates the Buffer Cache data
 * structures and finally removes the copy of the file in the Buffer Cache.
 *
 * @todo See if the error on tmpFS file delete can be handled in a better way.
 *
 * @param file Buffer Cache File Record for the file to be evicted
 */
static void
bc_evict_file(bc_file_t *file) {
    char path_tmpfs[PATH_MAX];
    struct stat statbuf;


    /* Validate the inputs */
    if ((file == NULL) || (bc_files.cached_files == 0)) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Identify the file in the buffer using the file path */
    if (!file->deleted) {
        dod(realpath(file->path, path_tmpfs) != NULL);

        /* Read the stats of the file */
        dod(!stat(file->path, &statbuf));

        /* Remove the symlink on the unshingled partition */
        dod(!unlink(file->path));

        /* Recreate the file */
        dod(!mknod(file->path, statbuf.st_mode, statbuf.st_dev));

        /* Remove the file from tmpFS */
        dod(!unlink(path_tmpfs));


        /* Update the file's metadata in the Attribute Replication file */
        xattr_set(file->path, X_STATS, &statbuf, sizeof(struct stat), 0);
    }


    /* Remove the element from the list and free it */
    bc_file_dequeue(file);
    free(file);
}



/**
 *
 * @brief Creates and initializes a new Cached File record
 *
 * @param path Path to the file that is being added to the buffer cache
 * @param inode Inode number of the file
 *
 * @return Pointer to a new Cached File record
 */
static bc_file_t
*bc_file_new(const char *path,
             int                 inode) {
    bc_file_t *new_file;


    /* Validate the inputs */
    if (path == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Create a new cached file record */
    dod((new_file = malloc(sizeof(bc_file_t))) != NULL);

    /* Initialize the new record and set its values */
    memset(new_file, 0, sizeof(bc_file_t));
    new_file->state = ST_LOADING;
    new_file->inode = inode;
    strcpy(new_file->path, path);
    new_file->fd = -1;
    new_file->fd_xattr = -1;
    new_file->deleted = 0;


    /* Initialize the synchronization primitives */
    dod(!pthread_mutex_init(&(new_file->mutex), NULL));
    dod(!pthread_cond_init(&(new_file->condvar), NULL));

    return new_file;
}



/**
 * @brief Enqueues a BC File record in the BC Files list
 *
 * This function DOES NOT perform any locking. It is the responsibility of the
 * caller to ensure that the correct locks are held before attempting to enqueue
 * a file.
 *
 * @param file Pointer to the BC File record to be enqueued
 * @param insert_at_head Boolean value, that causes insertion at the head or the
 * tail
 */
static void
bc_file_enqueue(bc_file_t       *file,
                int                      insert_at_head) {
    /* Validate the inputs */
    if (file == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Insert the record into the BC files list */
    if (bc_files.list == NULL) {
        bc_files.list = file;
        file->prev = file->next = file;
    } else {
        /* Insert the new element into the list just before the head */
        bc_files.list->prev->next = file;
        file->prev = bc_files.list->prev;
        file->next = bc_files.list;
        bc_files.list->prev = file;

        /* Reposition the list head, if needed */
        if (insert_at_head)
            bc_files.list = file;
    }

    /* Increment the record count */
    bc_files.cached_files++;
}



/**
 * @brief Dequeues a BC File record from the BC Files list
 *
 * This function DOES NOT perform any locking. It is the responsibility of the
 * caller to ensure that the correct locks are held before attempting to dequeue
 * a file.
 *
 * @param file Pointer to the BC File record to be dequeued
 */
static void
bc_file_dequeue(bc_file_t *file) {
    /* Validate the inputs */
    if (file == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Check if the list has only 1 record */
    if (bc_files.cached_files == 1) {
        /* Since there is only one node, clear the list */
        file->prev = NULL;
        file->next = NULL;
        bc_files.list = NULL;
    } else {
        /* Fix the pointers for the previous and next nodes */
        file->prev->next = file->next;
        file->next->prev = file->prev;

        /* Adjust the list head, if needed */
        if (file == bc_files.list)
            bc_files.list = file->next;
        /* Reset the file node pointers */
        file->prev = NULL;
        file->next = NULL;
    }

    /* Decrement the file count */
    bc_files.cached_files--;
}



/**
 * @brief Searches the BC File List using the file path
 *
 * This function DOES NOT perform any locking. It is the responsibility of the
 * caller to ensure that the correct locks are held before attempting to dequeue
 * a file.
 *
 * @param path Path of the file to be searched for
 *
 * @return Pointer to the BC File Record of the specified file
 */
bc_file_t
*bc_file_search_path(const char *path) {
    int i;
    bc_file_t *file;


    /* Validate the inputs */
    if (path == NULL) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }


    /* Iterate over the BC File List */
    for(i=0, file=bc_files.list; i<bc_files.cached_files; file=file->next, i++)
        if (!(strcmp(file->path, path)))
            return file;

    return NULL;
}

/**
 * @brief This function blocks until the file passed to it is processed in
 * the buffercache.
 *
 * In case the file is not dirty, this function returns immediately. In case
 * it is, it wait for the background worker thread to finish writing it to
 * disk before returning from this function.
 */
void bc_block_until_file_is_processed(bc_file_t *file_to_block_on) {
    dod(!pthread_mutex_lock(&bc_files.lock));
    dod(!pthread_mutex_lock(&(file_to_block_on->mutex)));
    if(file_to_block_on->dirty) {

	/* Mark the file state as STORING */
        file_to_block_on->state = ST_STORING;

        /* Reposition the file entry */
        bc_file_dequeue(file_to_block_on);
        bc_file_enqueue(file_to_block_on, 0);

        /* Enqueue a Request to write this file back to the EDI */
        log(LOG_BC_INFO, "Ref-count is 0; signaling worker to save file\n");
        dod(bc_enqueue_request(file_to_block_on, OP_STORE) != NULL);

        /* Signal to a worker to pick up the newly queued request */
        dod(!sem_post(&(bc_requests.count)));
	/* Increment the ref_count to force the worker to signal when
	 * the file has been stored. */
	file_to_block_on->ref_count++;

        /* Block waiting for the worker to finish writing the file */
        dod(!pthread_cond_wait(&file_to_block_on->condvar, &file_to_block_on->mutex));

	file_to_block_on->ref_count--;
	if(file_to_block_on->ref_count == 0) {
            file_to_block_on->state = ST_STORED;
	}
    }
    dod(!pthread_mutex_unlock(&(file_to_block_on->mutex)));
    dod(!pthread_mutex_unlock(&bc_files.lock));
}

/**
 * @brief Finds a file in the buffer cache that can be evicted safely
 *
 * If a file is not available immediately, i.e. if the file is being used by a
 * worker thread, then this function blocks waiting for the worker to finish its
 * task, and returns the file, after re-checking eviction conditions, should
 * they have changed while waiting for the worker to finish.
 *
 * This function DOES NOT perform any locking. It is the responsibility of the
 * caller to ensure that the correct locks are held before attempting to dequeue
 * a file.
 *
 * @return Pointer to the BC File record of the file to be evicted, if one is
 * found; NULL otherwise
 */
static bc_file_t
*bc_file_search_toevict() {
    int i;
    bc_file_t *cur_file;


    /* Iterate over the BC File list to find the record for the file */
    for(i=0, cur_file=bc_files.list; i<bc_files.cached_files;
        cur_file=cur_file->next, i++) {
        /* Lock the current file record */
        dod(!pthread_mutex_lock(&(cur_file->mutex)));

        /* Check the file's ref-count */
        if (cur_file->ref_count == 0) {
            /* Check if the file has been stored */
            if (cur_file->state == ST_STORED) {
                /* Return the current file as the eviction candidate */
                log(LOG_BC_INFO, "Found %s for eviction! State: %d, "
                    "Ref_count: %d\n", cur_file->path,
                    cur_file->state, cur_file->ref_count);
                return cur_file;

            } else if ((cur_file->state == ST_LOADED) ||
                       (cur_file->state == ST_STORING)) {
                /* Ensure that the file is LOADED or STORING. It could be
                 * possible that the file has a request for STORING in the queue
                 * that has not been processed yet. So the state could still be
                 * LOADED
                 *
                 * @todo Verify the above. State is being changed from LOADED to
                 * STORING synchronously in shingledfs_close(). So this might be
                 * an incorrect assumption!
                 */

                /* Increment the ref_count to force the worker to signal when
                 * the file has been stored. */
                cur_file->ref_count++;

                /* Block waiting for the worker to finish writing the file */
                dod(!pthread_cond_wait(&cur_file->condvar, &cur_file->mutex));

                /* Close the file */
                close(cur_file->fd);
                close(cur_file->fd_xattr);

                /* Decrement the ref_count to 0 */
                cur_file->ref_count = 0;
                cur_file->fd = -1;
                cur_file->fd_xattr = -1;
                cur_file->state = ST_STORED;

                /* Unlock the current file record */
                dod(!pthread_mutex_unlock(&(cur_file->mutex)));

                /* Return the current file as the eviction candidate */
                return cur_file;
            }
        }

        /* Unlock the current file record */
        dod(!pthread_mutex_unlock(&(cur_file->mutex)));
    }

    /* Return failure */
    return NULL;
}



/**
 * @brief Enqueues a BC Request
 *
 * @param file Pointer to the BC File record for the request
 * @param op The operation that is to be performed on the file
 *
 * @return Pointer to the newly created BC request
 */
static bc_request_t
*bc_enqueue_request(bc_file_t           *file,
                    bc_operation_t       op) {
    bc_request_t *request;


    /* Create a new cached file record */
    dod((request = malloc(sizeof(bc_request_t))) != NULL);

    /* Initialize the record */
    request->file = file;
    request->op = op;

    /* Lock the BC Request list */
    dod(!pthread_mutex_lock(&(bc_requests.lock)));

    /* Add the new request to the BC Request list */
    request->next = bc_requests.list;
    bc_requests.list = request;

    /* Unlock the BC Request list */
    dod(!pthread_mutex_unlock(&(bc_requests.lock)));

    /* Return the new request */
    return request;
}



/**
 * @brief Dequeues a BC Request
 *
 * This function returns the Request that is queued at the start of the BC
 * Request list, grabbing and releasing necessary locks to protect the shared
 * data structure
 *
 * @return Pointer to the dequeued Request struct, on success; NULL on failure
 */
static bc_request_t
*bc_dequeue_request() {
    bc_request_t *request = NULL;


    /* Lock the BC Request list */
    dod(!pthread_mutex_lock(&(bc_requests.lock)));

    /* Get the request at the head of the BC Request list */
    if (bc_requests.list != NULL) {
        request = bc_requests.list;
        bc_requests.list = bc_requests.list->next;
        request->next = NULL;
    }

    /* Unlock the BC Request list */
    dod(!pthread_mutex_unlock(&(bc_requests.lock)));

    /* Return the new request */
    return request;
}



/**
 * @brief Generates a file path on tmpFS for a file on the unshingled partition
 *
 * This function only fails when the output path exceeds PATH_MAX
 *
 * This function creates a unique path name for each tmpFS file using the
 * combination inodenum_filename. We assume that the inode number rollover is a
 * very very rare scenario and if it does happen, we will fail much earlier.
 *
 * @param path_out Buffer that will hold the generated path
 * @param path_in Path to the file on the unshingled partition
 * @param path_tmpfs Path to the tmpFS partition
 * @param path_unshingled Path to the unshingled partition
 */
void
bc_generate_tmpFS_path(char *path_out,
                       const char *path_in, int inode_num) {
    char *fname;

    /* Validate the inputs */
    if ((path_in == NULL) || (path_out == NULL)) {
        log(LOG_BC_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }

    /* Get only the name of the file from the path name */
    fname = strrchr(path_in, (int)'/');
    fname++;

    if ((strlen(SFSDATA.path_tmpfs) + strlen(fname)
         + INODE_MAX_STRLEN) >= PATH_MAX) {
        log(LOG_BC_ERRORS, "tmpFS for %s exceeds PATH_MAX!\n", path_in);
        exit(EXIT_FAILURE);
    }

    /* Construct the tmpFS file path */
    snprintf(path_out, PATH_MAX, "%s/%d_%s", SFSDATA.path_tmpfs,
             inode_num, fname);
    return;
}



/**
 * @brief Worker thread that handles data movement between the Buffer Cache and
 * the Emulated Disk Image (EDI)
 *
 * @todo Break this up in to multiple functions.
 */
static void
*bc_worker() {
    int retval;
    bc_request_t *request;
    char * buf;
    inode_t inode, temp_inode;
    size_t blocks, size, written;
    ssize_t ret;
    struct stat statbuf;
    open_band_t open_band;

#ifdef GENERATE_LOGS
    char filename[50];
    int len;
    time_t curtime = time(NULL);


    /* Initialize the logging interface */
    len = strftime(filename, 50, "/tmp/smr%Y%m%d%H%M%S_worker.log",
                   localtime(&curtime));
    if (!len) {
        fprintf(stderr, "Error generating log file name: %s\n", filename);
        assert(0);
    }
    log_open(filename);
#endif


    /* Log the call */
    log(LOG_WRK_INFO, "Initializing...\n");


    /* Allocate space for the buffer used to transfer data */
    size = (size_t)SFSDATA.edi_blksize * (size_t)SFSDATA.edi_bandsize;
#ifdef _EDI_ON_RAW
    dod(posix_memalign((void *)&buf, MEMALIGN_SIZE, size) == 0);
#else
    dod((buf = malloc(size)) != NULL);
#endif

    /* Loop forever */
    log(LOG_WRK_INFO, "Waiting for incoming requests...\n\n");
    while (1) {
        /* Wait for BC Requests */
        dod(!sem_wait(&bc_requests.count));


        /* Dequeue the request */
        if ((request = bc_dequeue_request()) == NULL) {
            log(LOG_WRK_ERRORS, "Error dequeuing request. No request found!\n");
            continue;
        }
        log(LOG_WRK_INFO, "New Request: Path: %s, Op: %d,  Inode No: %d, "
            "fd: %d, fd_xattr: %d, RefCount: %d, State: %d, "
            "Dirty: %d, Deleted: %d\n", request->file->path,
            request->op, request->file->inode, request->file->fd,
            request->file->fd_xattr, request->file->ref_count,
            request->file->state, request->file->dirty,
            request->file->deleted);


        /* Identify the type of request */
        switch(request->op) {
        case OP_LOAD:
            /* Read the stats of the file */
            dod(fgetxattr(request->file->fd_xattr, X_STATS, &statbuf,
                          sizeof(struct stat)) >= 0);

            /* Read the file inode */
            dod(!get_inode(request->file->inode, &inode));
            dod(inode.band_id > 0);
            dod(inode.size >= 0);

            /* Get the band to read */
            while((retval = band_meminfo_lock_band_for_read(inode.band_id)) == 0) {
               dod(!get_inode(request->file->inode, &temp_inode));
               if(inode.band_id == temp_inode.band_id) {
                   inode.band_offset = temp_inode.band_offset;
                   break;
               }
               band_meminfo_release_band_lock(inode.band_id);
               memcpy(&inode, &temp_inode, sizeof(inode));
            }
            /* TODO: Check error here */

            /* Read data from the disk */
#ifdef _EDI_ON_RAW
            dod(!edi_read(SFSDATA.fd_edi_metadata, SFSDATA.fd_edi, inode.band_id, inode.band_offset, buf,
                          inode.size));
#else
	    dod(!edi_read(SFSDATA.fd_edi, inode.band_id, inode.band_offset, buf, inode.size));
#endif

            retval = band_meminfo_release_band_lock(inode.band_id);
            /* TODO: Check error here */

            /* Write the data to the file */
            written = 0;
            size = statbuf.st_size;
            while ((ret = write(request->file->fd, buf + written,
                                size)) != size) {
                /* Ensure that the write didn't fail */
                dod(ret >= 0);

                /* Update the size and buffer offset */
                written += ret;
                size -= ret;
                log(LOG_WRK_INFO, "pwrite() returned %ld. Written: %ld. "
                    "Size remaining: %ld\n", ret, written, size);
            }


            /* Update the state of the record */
            dod(!pthread_mutex_lock(&(request->file->mutex)));
            request->file->state = ST_LOADED;
            dod(!pthread_mutex_unlock(&(request->file->mutex)));
            log(LOG_WRK_INFO, "Loaded %d bytes.Updated state to LOADED!\n",
                size);


            /* Awaken the threads waiting on the request */
            log(LOG_WRK_INFO, "Waking up thread(s) waiting on file\n");
            assert(!pthread_cond_broadcast(&(request->file->condvar)));
            break;

            /* Reset the file descriptor to be -1, since the operation failed */
            dod(!pthread_mutex_lock(&(request->file->mutex)));
            dod(!close(request->file->fd));
            dod(!close(request->file->fd_xattr));
            request->file->fd = -1;
            request->file->fd_xattr = -1;
            dod(!pthread_mutex_unlock(&(request->file->mutex)));

            break;


        case OP_STORE:
            /* Lock the File Record */
            dod(!pthread_mutex_lock(&(request->file->mutex)));
            /* Check if the file has been deleted */
            if (!request->file->deleted) {

                /* Check if the file is dirty */
                dod(request->file->dirty);
                log(LOG_WRK_INFO, "File is dirty!\n");

                /* Unlock the File Record */
                //dod(!pthread_mutex_unlock(&(request->file->mutex)));

                /* Calculate the no. of blocks to write */
                dod(!fstat(request->file->fd, &statbuf));
                blocks = statbuf.st_size;
           	blocks = (blocks + SFSDATA.edi_blksize - 1) / SFSDATA.edi_blksize;
            	log(LOG_WRK_INFO, "Need to write %d blocks!\n", blocks);

            	/* Retrieve an open band to write the file */
            	/* dod(get_openband(SFSDATA.fd_edi, &open_band, blocks) == 0); */
            	open_band.band = band_meminfo_get_free_band_for_write(blocks,
                                                    &open_band.rba, NULL);
            	if(open_band.band == 0) {
            	    log(LOG_WRK_INFO, "Not enough free space in any of the bands! Call Cleaning and try again later");
		    band_cleanall_wrapper(1);
                    open_band.band = band_meminfo_get_free_band_for_write(blocks,
                                                    &open_band.rba, NULL);
                    if(open_band.band == 0) {
            	    	log(LOG_WRK_INFO, "Not enough free space in any of the bands!!! Aborting");
                    	assert(0);
                    }
		    /* TODO: Handle this gracefully !! */
            	}

            	log(LOG_WRK_INFO, "Writing %d blocks (%ld bytes) to band %d at "
                	"rba %d retval was %d\n", blocks, statbuf.st_size,
                	open_band.band, open_band.rba, retval);
            	if (retval == 0) {
                    /* TODO: Bypass compiler error */
            	}

            	/* lock the inode table*/
                lock_inodelist();

		/* Read the file inode */
            	dod(!get_inode(request->file->inode, &inode));
            	log(LOG_WRK_INFO, "Retrieved file inode %d: Band: %d, RBA: %d, "
                	"Extent: %d, Inode No.: %d, Append Log Entry "
                	"No.: %d\n", request->file->inode, inode.band_id,
                	inode.band_offset, inode.size, inode.inode_num,
                	inode.entrynumber);

            	/* Mark any currently used band space by the file for cleaning */
            	if (inode.band_id > 0)
                    mark_for_cleaning(inode);

            	/* Calculate the no. of bytes to write */
            	dod(blocks <= SFSDATA.edi_bandsize);
            	size = SFSDATA.edi_blksize * blocks;
            	memset(buf + (SFSDATA.edi_blksize * (blocks - 1)), 0,
                   	SFSDATA.edi_blksize);

            	/* Read the file data */
            	dod(pread(request->file->fd, buf, size, 0) == statbuf.st_size);
            	log(LOG_WRK_INFO, "Read %ld bytes from file\n", statbuf.st_size);

            	/* Write the file data to the EDI */
#ifdef _EDI_ON_RAW
		retval = edi_write(SFSDATA.fd_edi_metadata, SFSDATA.fd_edi, open_band.band, open_band.rba, buf,
                               blocks);
#else
		retval = edi_write(SFSDATA.fd_edi, open_band.band, open_band.rba, buf, blocks);
#endif
            	assert(retval == 0);
            	log(LOG_WRK_INFO, "Wrote %ld bytes from tmpFS to EDI in Band %ld, "
                	"RBA %ld for %d blocks!\n", statbuf.st_size,
                	open_band.band, open_band.rba, blocks);

            	/* Update the inode */
            	inode_update(&inode, open_band.band, open_band.rba, blocks);

		/* release inode table lock*/
            	unlock_inodelist();

		retval = band_meminfo_release_band_lock(open_band.band);
            	/* TODO: Check error here */

            	/* Lock the File Record */
            	//dod(!pthread_mutex_lock(&(request->file->mutex)));
            }

            /* Check if the file was deleted while it was being written out */
            if (request->file->deleted) {

                /* Read the file inode */
                dod(!get_inode(request->file->inode, &inode));
                log(LOG_WRK_INFO, "Retrieved file inode %d: Band: %d, RBA: %d, "
                    "Extent: %d, Inode No.: %d, Append Log Entry "
                    "No.: %d\n", request->file->inode, inode.band_id,
                    inode.band_offset, inode.size, inode.inode_num,
                    inode.entrynumber);

                /* Unlink the inode */
                while( inode_unlink(request->file->inode, inode) != 0)
                {
                     /* Read the file inode */
                    dod(!get_inode(request->file->inode, &inode));
                    log(LOG_WRK_INFO, "REPEAT---Retrieved file inode %d: Band: %d, RBA: %d, "
                        "Extent: %d, Inode No.: %d, Append Log Entry "
                        "No.: %d\n", request->file->inode, inode.band_id,
                        inode.band_offset, inode.size, inode.inode_num,
                        inode.entrynumber);
                }

		inode_xattr_free(request->file->inode);
            }


            /* Mark the file as clean */
            request->file->dirty = 0;

            /* Copy the file's stats to disk */
            if (!request->file->deleted) {
                dod(!fsetxattr(request->file->fd_xattr, X_STATS, &statbuf,
                               sizeof(struct stat), 0));
            }

            /* Determine the cleanup strategy based on the ref-count */
            if (request->file->ref_count == 0) {
                /* Mark the state as STORED and close the open files */
                close(request->file->fd);
                close(request->file->fd_xattr);
                request->file->fd = -1;
                request->file->fd_xattr = -1;
                log(LOG_WRK_INFO, "Closed %s on tmpFS.\n", request->file->path);
                request->file->state = ST_STORED;

            } else {
                /* Mark the file state as LOADED and wake up any threads */
                request->file->state = ST_LOADED;
                dod(!pthread_cond_broadcast(&(request->file->condvar)));
                log(LOG_WRK_INFO, "Waking up thread(s) waiting on file\n");

            }

            /* Unlock the File Record */
            dod(!pthread_mutex_unlock(&(request->file->mutex)));

            break;


        default:
            /* Report the error for an unknown request type */
            log(LOG_WRK_ERRORS, "Unknown request type %d for request %s\n",
                request->op, request->file->path);
            dod(0);
            break;
        }


        /* Free the request struct */
        free(request);
        log(LOG_WRK_INFO, "Request completed!\n\n");
    }
}
