/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file buffercache.h
 *
 * @brief Macros, Type Definitions and Function Prototypes pertaining to the
 * Buffer Cache abstraction for ShingledFS
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#ifndef __BUFFERCACHE_H_
#define __BUFFERCACHE_H_
#ifdef __cplusplus
extern "C" {
#endif

#include <pthread.h>



/* ====================== */
/* -- Type Definitions -- */
/* ====================== */

/**
 * @brief Defines the state of a Cached File record
 */
typedef enum {
    ST_LOADING,
    ST_LOADED,
    ST_STORING,
    ST_STORED
}bc_state_t;


/**
 * @brief Defines an operation performed by the Buffer Cache
 */
typedef enum {
    OP_LOAD,
    OP_STORE
}bc_operation_t;


/**
 * @brief Defines a Cached File record
 *
 * @todo Hide any items that do not need to be exposed to FUSE. Currently, the
 * only thing that FUSE needs is the file descriptor and the dirty flag.
 * However, its better to store a pointer to this struct  in FUSE for constant-
 * order searching of the record for a file. So, move all other fields into
 * another struct and save a void pointer to that struct within this struct.
 * This way only the Buffer Cache code can access the internal fields, whereas
 * FUSE can use the values that it is concerned with.
 *
 * A thread will lock the structure by using the built-in mutex, increment the
 * ref-count on open() or decrement it on close(), and pass the record to a
 * worker thread to load/unload the file data from/to the EDI.
 *
 * Should the file be in a loading state, a thread can increment/decrement the
 * ref-count as applicable, and wait on the condition variable for the worker
 * thread to wake it up once the transfer is complete.
 *
 * Worker threads will load/save the file from/to the EDI, lock the mutex, and
 * check the ref-count. If the ref-count is greater than 0, then the worker will
 * broadcast on the condition variable to awaken any threads waiting for the
 * transfer to finish.
 *
 * @param path The path to the open file
 * @param inode The inode number of the file
 * @param stats The attributes of the file, cached in memory for fast access
 * @param fd The file descriptor assigned to the open file
 * @param ref_count The reference count for the open file
 * @param dirty Flag describing whether the file was written to since the load
 * @param state Current state of the file
 * @param mutex The mutex protecting the open_file record
 * @param cond_var The condition variable used to wait on the open_file record
 * @param prev Pointer to the previous open file node
 * @param next Pointer to the next open file node
 */
typedef struct bc_file{
    char                                 path[PATH_MAX];
    int                                  inode;
    int                                  fd;
    int                                  fd_xattr;
    int                                  deleted;
    int                                  ref_count;
    int                                  dirty;
    bc_state_t                   state;

    pthread_mutex_t              mutex;
    pthread_cond_t               condvar;

    struct bc_file              *prev;
    struct bc_file              *next;
}bc_file_t;


/**
 * @brief Defines a request placed by ShingledFS to load/store a file
 *
 * @param file Pointer to the BC File Record that holds the file information
 * @param op The operation to perform: load or store.
 * @param next Pointer to the next BC Request Record
 */
typedef struct bc_request{
    bc_file_t                   *file;
    bc_operation_t               op;

    struct bc_request   *next;
} bc_request_t;



/* ========================= */
/* -- Function Prototypes -- */
/* ========================= */

void             bc_init(size_t cache_size);
bc_file_t       *bc_open(const char *path);
int              bc_destory();
void             bc_close(bc_file_t *file);
int              bc_unlink(const char *path);
int              bc_rename(const char *path, const char *path_new);
void             bc_generate_tmpFS_path(char *path_out, const char *path_in, int inode_num);
bc_file_t       *bc_file_search_path(const char *path);
void             bc_clean();
void 		 bc_block_until_file_is_processed(bc_file_t *file_to_block_on);


#ifdef __cplusplus
}
#endif
#endif /* __BUFFERCACHE_H_ */
