/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <assert.h>

#include "band_log.h"
#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <semaphore.h>
#include <pthread.h>
#include <stdlib.h>
#include <time.h>

#include "band_bitmap.h"
#include "cleaning_thread.h"
#include "common.h"
#include "inode.h"
#include "emulator.h"
#include "stats.h"
#include "band_meminfo.h"
#include "helper.h"

/*
 * Cleaner Tunables
 */
/**
 * The minimum interval between two cleaning cycles. Cleaner
 * wakes up every CLEANER_INTERVAL_SECS tocheck if it has to
 * clean */
#define CLEANER_INTERVAL_SECS 10

/**
 * Threshold for amount of idle time after which we declare
 * that the system is idle
 */
#define FS_IDLE_THRESHOLD_SECS 10

/**
 * @brief Decides if the system is idle
 *
 * Used by the cleaner thread to decide if the system
 * is idle and then kickoff cleaning. This fuction
 * should look at various stats to decide the idleness.
 *
 * @param None
 * @return 1 - system is idle
 * @return 0 - system is not idle
 */
int is_sys_idle(void)
{
    /* Test algo : free once in 5 intervals */
    /*static int times_called;
    if (++times_called == 5) {
        times_called = 0;
        return 1;
    }
    */

    /* Free if a predefined time has elapsed since
     * a read/write happened in the system */
    if ((time(NULL) - stats_get(LAST_ACTIVITY_TIME)) > FS_IDLE_THRESHOLD_SECS) {
        return 1;
    }

    return 0;

}
/*
 * @brief : Spawns a cleaning thread and intializes the cleaning parameters.
 *
 * @param : None.
 *
 * @return 0 : on Success.
 *         -1 : on Error.
 *
 */

int cleaner_init(void)
{
    int err;
    pthread_t tid;

    if ((err = pthread_create(&tid, NULL, band_cleaner, NULL)) != 0) {
        log(LOG_CLN_ERRORS, "Error creating thread: %s\n", strerror(err));
        return -1;
    }

    return 0;
}



/*
 * @brief : Cleaning thread created by the cleaner_init function. It waits on
 *          a semaphore to be signalled by the signal handler to start
 *          cleaning. Currently ascynchronus cleaning is supported. It will
 *          clean all bands one by one and then will again wait on the semaphore
 *          once all the bands are cleaned. It also intializes the log
 *          functions for the cleaning thread.
 *
 * @param sfs : Pointer to fuse context data.
 *
 * @return : None.
 *
 */

void * band_cleaner()
{
#ifdef GENERATE_LOGS
    char filename[50];
    int len;
    time_t curtime = time(NULL);

    /* Initialize the logging interface */
    len = strftime(filename, 50, "/tmp/smr%Y%m%d%H%M%S_cleaner.log",
                   localtime(&curtime));
    if (!len) {
        fprintf(stderr, "Error generating log file name: %s\n", filename);
        assert(0);
    }
    log_open(filename);
#endif

    log(LOG_CLN_ALL, "Started cleaner thread...\n");

    struct timespec snooze;
    snooze.tv_sec = CLEANER_INTERVAL_SECS;
    snooze.tv_nsec = 0;

    while (1) {
        int fs_full = stats_disk_utilization();
        /* periodically check if the system is idle (opportunistic)
         * OR if it is absolutely necessary to clean (threshold)*/
        log(LOG_CLN_ALL, "disk utilization %d\n", fs_full);
        if (fs_full >= FS_FULL_THRESHOLD_SOFT) {
            log(LOG_CLN_ALL, "disk utilization %d > soft %d\n", fs_full,
                    FS_FULL_THRESHOLD_SOFT);

            if (fs_full >= FS_FULL_THRESHOLD_HARD ||
                is_sys_idle()) {
                log(LOG_CLN_ALL, "disk utilization %d > hard %d\n", fs_full,
                    FS_FULL_THRESHOLD_HARD);
                /* will clean some bands until utilization drops below SOFT */
                band_cleanall_wrapper(0);
            }
        }

        nanosleep(&snooze, NULL);

        /* Refresh the stats. Reusing the cleaner thread instead
         * of running a seperate stats thread */
        stats_update();
    }
}

/*
 * @brief : Fucntion calls band_cleanall. Add wrapper to collect cleaning stats
 *
 * @param num_bands_to_clean : How many bands to clean in this run
 *
 * @return : void
 *
 */


void band_cleanall_wrapper(int num_bands_to_clean) {
    char cleaner[2048];
    int clean_flag = 0;
    stats_cleaner(BEFORE, 0, cleaner);
    clean_flag = band_cleanall(num_bands_to_clean);
    stats_cleaner(AFTER, clean_flag, cleaner);
}

/*
 * @brief : Function will clean all the band on the disk. Will return once all
 *          the bands are cleaned. When this function is running no other
 *          activity should be running on the file system.
 *
 * @param sfs : Pointer to fuse context.
 * @param num_bands_to_clean : How many bands to clean in this run
 *
 * @return : 1 if cleaning happens.
 *           0 if no cleaning.
 *
 */

int band_cleanall(int num_bands_to_clean)
{
    int n = 0;
    time_t t = time(NULL);
    log(LOG_CLN_INFO, "Cleaning start time %s\n", ctime(&t));
    while (band_cleaning() != -1) {
        n++;
        if (time_to_stop(n, num_bands_to_clean, FIXED_THRESHOLD))
            break;
    }
    log(LOG_CLN_INFO, "Cleaned %d bands in this run\n", n);
    if(n == 0)
	return 0;
    else
        return 1;
}


/*
 * @brief : Cleans one band at a time.
 *          Selects one band for cleaning and one band to clean to.
 *          Once the band are selected according to the policies then they
 *          are cleaned. The updates of the data and inode takes place in
 *          a particular order to avoid inconsistencies during failure.
 *
 * @param
 *
 * @return 0 : on Success.
 *         -1 : no band left to clean.
 *
 */

int band_cleaning()
{
    band_t band_frag;
    band_t band_clean;
    band_clean_t band_clean_tuple;
    struct store_entry *live;
    int length;
    int write_pointer = 0;
    //if (select_cleaning_band(&band_clean_tuple,
    //            COST_BENEFIT) != 0) {
    //if (select_cleaning_band(&band_clean_tuple,
    //            GET_BAND_TO_CLEAN_FIRST) != 0) {
    if (select_cleaning_band(&band_clean_tuple, GREEDY_BY_MAX_SPACE_RECOVERED) != 0) {
        return -1;
    }
    band_frag = band_clean_tuple.band_to_clean;
    band_clean = band_clean_tuple.band_clean_to;
    get_livelog_wrapper(band_frag, &live, &length);

    if (band_frag == 0) {
        assert(0);
        return -1;
    }
    if (band_clean == 0) {
        //in place cleaning policy
        //migrate data from band_to_clean to cleaning buffer band
        //migrate data from cleaning buffer band back to band_to_clean
        log(LOG_CLN_INFO, "Cleaning band %d (%d fragments) in place using band 0\n",
            band_frag, length);
#ifdef _EDI_ON_RAW
        if (edi_managebands(SFSDATA.fd_edi_metadata, SVC_SETPTR, &band_clean, &write_pointer,
                        sizeof(rba_t))) {
#else
	if (edi_managebands(SFSDATA.fd_edi, SVC_SETPTR, &band_clean, &write_pointer, sizeof(rba_t))) {
#endif
	    assert(0);
            exit(EXIT_FAILURE);
        }
        migrate_data(band_frag, 0, live, length);
        free(live);
#ifdef _EDI_ON_RAW
        if (edi_managebands(SFSDATA.fd_edi_metadata, SVC_SETPTR, &band_frag, &write_pointer,
                        sizeof(rba_t))) {
#else
	if (edi_managebands(SFSDATA.fd_edi, SVC_SETPTR, &band_frag, &write_pointer, sizeof(rba_t))) {
#endif
            assert(0);
            exit(EXIT_FAILURE);
        }
        get_livelog_wrapper(0, &live,  &length);
        migrate_data(0, band_frag, live, length);
        //only need to release one lock
        band_meminfo_release_band_lock(band_frag);
    } else {
        log(LOG_CLN_INFO, "Cleaning band %d (%d fragments) into clean band %d\n",
            band_frag, length, band_clean);

        migrate_data(band_frag, band_clean, live, length);
#ifdef _EDI_ON_RAW
        assert (edi_managebands(SFSDATA.fd_edi_metadata, SVC_SETPTR, &band_frag, &write_pointer,
                        sizeof(rba_t)) == 0);
#else
	assert (edi_managebands(SFSDATA.fd_edi, SVC_SETPTR, &band_frag, &write_pointer, sizeof(rba_t)) == 0);
#endif

        //release locks for both the bands.
        band_meminfo_release_band_lock(band_clean_tuple.band_clean_to);
        band_meminfo_release_band_lock(band_clean_tuple.band_to_clean);
    }

    free(live);
    return 0;
}


/*
 * @brief : The function migrates data from one band to another band.
 *
 * @param band_to_clean : Band id of the band to be cleaned.
 * @param band_clean_to : Band id of the band to clean to.
 * @param live : Array of store entries i.e. live data's metadata for the band.
 * @param length : Length of the array live.
 *
 * @return : None.
 *
 */

void migrate_data(band_t band_to_clean, band_t band_clean_to,
                  struct store_entry *live, int length)
{
    int i;
    int write_pointer;
    inode_t temp_inode;
    int index;
    int clean_storefd;
    int err;
    size_t cleaning_size = 0;

#ifdef _EDI_ON_RAW
    if (edi_managebands(SFSDATA.fd_edi_metadata, SVC_GETPTR, &band_clean_to, &write_pointer,
                        sizeof(rba_t))) {
#else
    if (edi_managebands(SFSDATA.fd_edi, SVC_GETPTR, &band_clean_to, &write_pointer, sizeof(rba_t))) {
#endif
	assert(0);
        exit(EXIT_FAILURE);
    }

    clean_storefd = init_storelog(band_clean_to);
    if (clean_storefd == -1) {
	assert(0);
        exit(EXIT_FAILURE);
    }

    /* lock inode table*/
    dbg("migrate data %d to %d lock inodelist\n", band_to_clean, band_clean_to);
    lock_inodelist();
    for(i = 0; i < length; i++) {

        err = get_inode(live[i].inode_num, &temp_inode);
        if (err == -1) {
            /*release inode table lock*/
            unlock_inodelist();
	    assert(0);
            exit(EXIT_FAILURE);
        }

	copy_data(band_to_clean, band_clean_to, live[i].offset, live[i].size,
                  write_pointer);
        //sum of live[i].size for update in memory data structure
        cleaning_size += live[i].size;

        index = append_storelog(band_clean_to, clean_storefd, write_pointer, live[i].size,
                                live[i].inode_num, 0);

        temp_inode.entrynumber = index;
        temp_inode.band_id = band_clean_to;
        temp_inode.band_offset = write_pointer;

        err = put_inode(live[i].inode_num, &temp_inode);

        if (err == -1) {
            /*release inode table lock*/
            unlock_inodelist();
            assert(0);
	    exit(EXIT_FAILURE);
        }

        write_pointer = write_pointer + live[i].size;

    }

    /*release inode table lock*/
    unlock_inodelist();
    band_meminfo_update_band_cleaned(band_to_clean);
    clear_storelog(band_to_clean);
    clear_deletelog(band_to_clean);
    close(clean_storefd);
}

/*
 * @brief : Copy data from one band to another band with the given band offset
 *          and size.
 *
 * @param band_src : Source band number.
 * @param band_dst : Destination band number.
 * @param offset_src : Sorce band rba.
 * @param len_src : Length of copy.
 * @param offset_dst : Destination band rba.
 *
 * @return None.
 *
 */

void copy_data(int band_src, int band_dst, int offset_src, int len_src,
               int offset_dst)
{
    char *buf;
    assert(posix_memalign((void *)&buf, MEMALIGN_SIZE, (len_src * SFSDATA.edi_blksize)) == 0);

    if (buf == NULL) {
        assert(0);
	exit(EXIT_FAILURE);
    }

#ifdef _EDI_ON_RAW
    if (edi_read(SFSDATA.fd_edi_metadata, SFSDATA.fd_edi, band_src, offset_src, buf, len_src)) {
#else
    if (edi_read(SFSDATA.fd_edi, band_src, offset_src, buf, len_src)) {
#endif
        exit(EXIT_FAILURE);
    }

#ifdef _EDI_ON_RAW
    if (edi_write(SFSDATA.fd_edi_metadata, SFSDATA.fd_edi, band_dst, offset_dst, buf, len_src)) {
#else
    if (edi_write(SFSDATA.fd_edi, band_dst, offset_dst, buf, len_src)) {
#endif
        exit(EXIT_FAILURE);
    }

}
