/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file profiler.c
 *
 * @brief Builds a profile of the FUSE functions called by a workload.
 *
 * The majority of the FUSE code has been copied from fusequota, downloaded
 * from http://code.google.com/p/fusequota
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#define FUSE_USE_VERSION 26

#define _XOPEN_SOURCE 500

#include <fuse.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/xattr.h>
#include <sys/types.h>

#include <semaphore.h>
#include <pthread.h>
#include <signal.h>



/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

#define GETATTR      0
#define READLINK     1
#define MKNOD        2
#define MKDIR        3
#define UNLINK       4
#define RMDIR        5
#define SYMLINK      6
#define RENAME       7
#define LINK         8
#define CHMOD        9
#define CHOWN       10
#define TRUNCATE    11
#define UTIME       12
#define OPEN        13
#define READ        14
#define WRITE       15
#define STATFS      16
#define RELEASE     17
#define FSYNC       18
#define SETXATTR    19
#define GETXATTR    20
#define LISTXATTR   21
#define REMOVEXATTR 22
#define OPENDIR     23
#define READDIR     24
#define RELEASEDIR  25
#define ACCESS      26

static int counters[27] = {0};
static FILE *file;



/* ====================== */
/* -- Static Variables -- */
/* ====================== */

static char base[PATH_MAX];
static sem_t sem_profiler;



/* ============================== */
/* -- Function Implementations -- */
/* ============================== */


void *profiler() {
    while(1) {
        /* Wait for a profiling signal */
        sem_wait(&sem_profiler);


        /* Open the file to write the application's profile */
        file = fopen("/tmp/profile.out", "a+");
        fprintf(file, "Profile:\n");
        fprintf(file, "getattr()\t%d\n", counters[GETATTR]);
        fprintf(file, "readlink()\t%d\n", counters[READLINK]);
        fprintf(file, "mknod()\t%d\n", counters[MKNOD]);
        fprintf(file, "mkdir()\t%d\n", counters[MKDIR]);
        fprintf(file, "unlink()\t%d\n", counters[UNLINK]);
        fprintf(file, "rmdir()\t%d\n", counters[RMDIR]);
        fprintf(file, "symlink()\t%d\n", counters[SYMLINK]);
        fprintf(file, "rename()\t%d\n", counters[RENAME]);
        fprintf(file, "link()\t%d\n", counters[LINK]);
        fprintf(file, "chmod()\t%d\n", counters[CHMOD]);
        fprintf(file, "chown()\t%d\n", counters[CHOWN]);
        fprintf(file, "truncate()\t%d\n", counters[TRUNCATE]);
        fprintf(file, "utime()\t%d\n", counters[UTIME]);
        fprintf(file, "open()\t%d\n", counters[OPEN]);
        fprintf(file, "read()\t%d\n", counters[READ]);
        fprintf(file, "write()\t%d\n", counters[WRITE]);
        fprintf(file, "statfs()\t%d\n", counters[STATFS]);
        fprintf(file, "release()\t%d\n", counters[RELEASE]);
        fprintf(file, "fsync()\t%d\n", counters[FSYNC]);
        fprintf(file, "setxattr()\t%d\n", counters[SETXATTR]);
        fprintf(file, "getxattr()\t%d\n", counters[GETXATTR]);
        fprintf(file, "listxattr()\t%d\n", counters[LISTXATTR]);
        fprintf(file, "removexattr()\t%d\n", counters[REMOVEXATTR]);
        fprintf(file, "opendir()\t%d\n", counters[OPENDIR]);
        fprintf(file, "readdir()\t%d\n", counters[READDIR]);
        fprintf(file, "releasedir()\t%d\n", counters[RELEASEDIR]);
        fprintf(file, "access()\t%d\n", counters[ACCESS]);
        fprintf(file, "\n\n");
        fclose(file);
    }

    return NULL;
}



void sig_handler(int sig) {
    /* Check the signal received */
    if (sig == SIGUSR1) {
        fprintf(stderr, "Signaling...\n");
        sem_post(&sem_profiler);
        signal(sig, sig_handler);
    }

}



/**
 * Appends the path of the root filesystem to the given path, returning
 * the result in buf.
 */
void
fullpath (const char *path, char *buf)
{
    char *basedir = (char *) fuse_get_context ()->private_data;

    strcpy (buf, basedir);
    strcat (buf, path);
}


/* The following functions describe FUSE operations. Each operation appends
   the path of the root filesystem to the given path in order to give the
   mirrored path. All quota checking takes place in fuse_write. */

int
fuse_getattr (const char *path, struct stat *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[GETATTR]++;
    retval = lstat (fpath, buf) ? -errno : 0;
    return retval;
}

int
fuse_readlink (const char *path, char *target, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[READLINK]++;
    retval = readlink (fpath, target, size);
    if (retval > 0) {
        if (retval == size)
            retval--;
        target[retval] = '\0';
        retval = 0;
    }
    return retval;
}

int
fuse_mknod (const char *path, mode_t mode, dev_t dev)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[MKNOD]++;
    retval = mknod (fpath, mode | S_IRUSR | S_IWUSR, dev) ? -errno : 0;
    return retval;
}

int
fuse_mkdir (const char *path, mode_t mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[MKDIR]++;
    retval = mkdir (fpath, mode | S_IFDIR) ? -errno : 0;
    return retval;
}

int
fuse_unlink (const char *path)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[UNLINK]++;
    retval = unlink (fpath) ? -errno : 0;
    return retval;
}

int
fuse_rmdir (const char *path)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[RMDIR]++;
    retval = rmdir (fpath) ? -errno : 0;
    return retval;
}

int
fuse_symlink (const char *path, const char *link)
{
    char flink[PATH_MAX];
    int retval;
    fullpath (link, flink);

    counters[SYMLINK]++;
    retval = symlink (path, flink) ? -errno : 0;
    return retval;
}

int
fuse_rename (const char *src, const char *dst)
{
    char fsrc[PATH_MAX];
    char fdst[PATH_MAX];
    int retval;
    fullpath (src, fsrc);
    fullpath (dst, fdst);

    counters[RENAME]++;
    retval = rename (fsrc, fdst) ? -errno : 0;
    return retval;
}

int
fuse_link (const char *src, const char *dst)
{
    char fsrc[PATH_MAX];
    char fdst[PATH_MAX];
    int retval;
    fullpath (src, fsrc);
    fullpath (dst, fdst);

    counters[LINK]++;
    retval = link (fsrc, fdst) ? -errno : 0;
    return retval;
}

int
fuse_chmod (const char *path, mode_t mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[CHMOD]++;
    retval = chmod (fpath, mode) ? -errno : 0;
    return retval;
}

int
fuse_chown (const char *path, uid_t uid, gid_t gid)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[CHOWN]++;
    retval = chown (fpath, uid, gid) ? -errno : 0;
    return retval;
}

int
fuse_truncate (const char *path, off_t off)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[TRUNCATE]++;
    retval = truncate (fpath, off) ? -errno : 0;
    return retval;
}

int
fuse_utime (const char *path, struct utimbuf *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[UTIME]++;
    retval = utime (fpath, buf) ? -errno : 0;
    return retval;
}

int
fuse_open (const char *path, struct fuse_file_info *fi)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[OPEN]++;
    int fh = open (fpath, fi->flags);
    if (fh < 0)
        retval = -errno;
    else {
        fi->fh = fh;
        retval = 0;
    }

    return retval;
}

int
fuse_read (const char *path, char *buf, size_t size, off_t off,
           struct fuse_file_info *fi)
{
    int retval;

    counters[READ]++;
    retval = pread (fi->fh, buf, size, off) < 0 ? -errno : size;
    return retval;
}

int
fuse_write (const char *path, const char *buf, size_t size, off_t off,
            struct fuse_file_info *fi)
{
    int retval;

    counters[WRITE]++;
    retval = pwrite (fi->fh, buf, size, off) < 0 ? -errno : size;
    return retval;
}

int
fuse_statfs (const char *path, struct statvfs *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[STATFS]++;
    retval = statvfs (fpath, buf) ? -errno : 0;
    return retval;
}

int
fuse_release (const char *path, struct fuse_file_info *fi)
{
    int retval;

    counters[RELEASE]++;
    retval = close (fi->fh) ? -errno : 0;
    return retval;
}

int
fuse_fsync (const char *path, int datasync, struct fuse_file_info *fi)
{
    int retval;

    counters[FSYNC]++;
    if (datasync)
        retval = fdatasync (fi->fh) ? -errno : 0;
    else
        retval = fsync (fi->fh) ? -errno : 0;

    return retval;
}

int
fuse_setxattr (const char *path, const char *name, const char *value,
               size_t size, int flags)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[SETXATTR]++;
    retval = lsetxattr (fpath, name, value, size, flags) ? -errno : 0;
    return retval;
}

int
fuse_getxattr (const char *path, const char *name, char *value, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[GETXATTR]++;
    ssize_t s = lgetxattr (fpath, name, value, size);
    retval = s < 0 ? -errno : s;
    return retval;
}

int
fuse_listxattr (const char *path, char *list, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[LISTXATTR]++;
    retval = llistxattr (fpath, list, size) < 0 ? -errno : 0;
    return retval;
}

int
fuse_removexattr (const char *path, const char *name)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[GETATTR]++;
    retval = lremovexattr (fpath, name) ? -errno : 0;
    return retval;
}

int
fuse_opendir (const char *path, struct fuse_file_info *fi)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[OPENDIR]++;
    DIR *dir = opendir (fpath);
    if (dir == NULL)
        retval = -errno;
    else {
        fi->fh = (uint64_t) dir;
        retval = 0;
    }

    return retval;
}

int
fuse_readdir (const char *path, void *buf, fuse_fill_dir_t fill, off_t off,
              struct fuse_file_info *fi)
{
    struct dirent *de = NULL;

    counters[READDIR]++;
    while ((de = readdir ((DIR *) fi->fh)) != NULL)
    {
        struct stat st;
        memset (&st, 0, sizeof (struct stat));
        st.st_ino = de->d_ino;
        st.st_mode = de->d_type << 12;

        if (fill (buf, de->d_name, &st, 0))
            break;
    }

    return 0;
}

int
fuse_releasedir (const char *path, struct fuse_file_info *fi)
{
    int retval;

    counters[RELEASEDIR]++;
    retval = closedir ((DIR *) fi->fh) ? -errno : 0;
    return retval;
}

int
fuse_access (const char *path, int mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    counters[ACCESS]++;
    retval = access (fpath, mode) ? -errno : 0;
    return retval;
}

void *
fuse_init (struct fuse_conn_info *conn)
{
    pthread_t tid;
    int err;

    /* Initialize the semaphore */
    if ((err = sem_init(&sem_profiler, 0, 0)) != 0)
        exit(EXIT_FAILURE);

    /* Install the signal handler */
    signal(SIGUSR1, sig_handler);

    /* Spawn a thread to profile the application */
    if ((err = pthread_create(&tid, NULL, profiler, NULL)) != 0)
        exit(EXIT_FAILURE);

    return (fuse_get_context())->private_data;
}



struct fuse_operations fuse_ops = {
    .getattr = fuse_getattr,
    .readlink = fuse_readlink,
    .mknod = fuse_mknod,
    .mkdir = fuse_mkdir,
    .unlink = fuse_unlink,
    .rmdir = fuse_rmdir,
    .symlink = fuse_symlink,
    .rename = fuse_rename,
    .link = fuse_link,
    .chmod = fuse_chmod,
    .chown = fuse_chown,
    .truncate = fuse_truncate,
    .utime = fuse_utime,
    .open = fuse_open,
    .read = fuse_read,
    .write = fuse_write,
    .statfs = fuse_statfs,
    .release = fuse_release,
    .fsync = fuse_fsync,
    .setxattr = fuse_setxattr,
    .getxattr = fuse_getxattr,
    .listxattr = fuse_listxattr,
    .removexattr = fuse_removexattr,
    .opendir = fuse_opendir,
    .readdir = fuse_readdir,
    .releasedir = fuse_releasedir,
    .access = fuse_access,
    .init = fuse_init,
};



void
usage ()
{
    printf ("transparency <basedir> <mountpoint>\n");
    exit (0);
}



int
main (int argc, char *argv[])
{
    if (argc < 3)
        usage ();

    if (realpath(argv[1], base) == NULL)
        printf("The base directory path %s is invalid!\n", argv[1]);

    int i = 1;
    for (; i < argc; i++)
        argv[i] = argv[i + 1];
    argc -= 1;

    umask(0);
    int ret = fuse_main (argc, argv, &fuse_ops, base);
    if (ret < 0)
        printf("Error in FUSE main()\n");
    else
        printf("Output will be found in /tmp/profile.out\n");

    return ret;
}
