/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file transparentFuse.c
 *
 * @brief Transparent FUSE program.
 *
 * The majority of the FUSE code has been copied from fusequota, downloaded
 * from http://code.google.com/p/fusequota
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * @date Jan 20, 2012 2:35:31 PM
 */

#define FUSE_USE_VERSION 26

#define _XOPEN_SOURCE 500

#include <fuse.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/xattr.h>
#include <sys/types.h>



FILE *file_log;
#ifdef GENERATE_LOGS
#define log(...)                                        \
    do {                                                \
        fprintf(file_log, "%20s: ", __FUNCTION__);      \
        fprintf(file_log, __VA_ARGS__);                 \
    } while(0)
#else
#define log(...)
#endif


char base[PATH_MAX];


/**
 * Appends the path of the root filesystem to the given path, returning
 * the result in buf.
 */
void
fullpath (const char *path, char *buf)
{
    char *basedir = (char *) fuse_get_context ()->private_data;

    strcpy (buf, basedir);
    strcat (buf, path);
}


/* The following functions describe FUSE operations. Each operation appends
   the path of the root filesystem to the given path in order to give the
   mirrored path. All quota checking takes place in fuse_write. */

int
fuse_getattr (const char *path, struct stat *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = lstat (fpath, buf) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_readlink (const char *path, char *target, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, target=%s, size=%ld\n", path, target, size);
    retval = readlink (fpath, target, size);
    if (retval > 0) {
        if (retval == size)
            retval--;
        target[retval] = '\0';
        retval = 0;
    }
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_mknod (const char *path, mode_t mode, dev_t dev)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, mode=0%o, dev=%ld\n", path, mode, dev);
    retval = mknod (fpath, mode | S_IRUSR | S_IWUSR, dev) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_mkdir (const char *path, mode_t mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, mode=0%o\n", path, mode);
    retval = mkdir (fpath, mode | S_IFDIR) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_unlink (const char *path)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = unlink (fpath) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_rmdir (const char *path)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = rmdir (fpath) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_symlink (const char *path, const char *link)
{
    char flink[PATH_MAX];
    int retval;
    fullpath (link, flink);

    log("path=%s, link=%s\n", path, link);
    log("path=%s, flink=%s\n", path, flink);
    retval = symlink (path, flink) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_rename (const char *src, const char *dst)
{
    char fsrc[PATH_MAX];
    char fdst[PATH_MAX];
    int retval;
    fullpath (src, fsrc);
    fullpath (dst, fdst);

    log("src=%s, dst=%s\n", src, dst);
    retval = rename (fsrc, fdst) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_link (const char *src, const char *dst)
{
    char fsrc[PATH_MAX];
    char fdst[PATH_MAX];
    int retval;
    fullpath (src, fsrc);
    fullpath (dst, fdst);

    log("src=%s, dst=%s\n", src, dst);
    retval = link (fsrc, fdst) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_chmod (const char *path, mode_t mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, mode=0%o\n", path, mode);
    retval = chmod (fpath, mode) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_chown (const char *path, uid_t uid, gid_t gid)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, uid=%d, gid=%d\n", path, uid, gid);
    retval = chown (fpath, uid, gid) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_truncate (const char *path, off_t off)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, off=%ld\n", path, off);
    retval = truncate (fpath, off) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_utime (const char *path, struct utimbuf *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = utime (fpath, buf) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_open (const char *path, struct fuse_file_info *fi)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s, flags=0%o\n", path, fi->flags);
    int fh = open (fpath, fi->flags);
    if (fh < 0)
        retval = -errno;
    else {
        fi->fh = fh;
        retval = 0;
    }

    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_read (const char *path, char *buf, size_t size, off_t off,
           struct fuse_file_info *fi)
{
    int retval;

    log("path=%s\n", path);
    retval = pread (fi->fh, buf, size, off) < 0 ? -errno : size;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_write (const char *path, const char *buf, size_t size, off_t off,
            struct fuse_file_info *fi)
{
    int retval;

    log("path=%s\n", path);
    retval = pwrite (fi->fh, buf, size, off) < 0 ? -errno : size;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_statfs (const char *path, struct statvfs *buf)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = statvfs (fpath, buf) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_release (const char *path, struct fuse_file_info *fi)
{
    int retval;

    log("path=%s\n", path);
    retval = close (fi->fh) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_fsync (const char *path, int datasync, struct fuse_file_info *fi)
{
    int retval;

    log("path=%s\n", path);
    if (datasync)
        retval = fdatasync (fi->fh) ? -errno : 0;
    else
        retval = fsync (fi->fh) ? -errno : 0;

    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_setxattr (const char *path, const char *name, const char *value,
               size_t size, int flags)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = lsetxattr (fpath, name, value, size, flags) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_getxattr (const char *path, const char *name, char *value, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    ssize_t s = lgetxattr (fpath, name, value, size);
    retval = s < 0 ? -errno : s;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_listxattr (const char *path, char *list, size_t size)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = llistxattr (fpath, list, size) < 0 ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_removexattr (const char *path, const char *name)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = lremovexattr (fpath, name) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_opendir (const char *path, struct fuse_file_info *fi)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    DIR *dir = opendir (fpath);
    if (dir == NULL)
        retval = -errno;
    else {
        fi->fh = (uint64_t) dir;
        retval = 0;
    }

    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_readdir (const char *path, void *buf, fuse_fill_dir_t fill, off_t off,
              struct fuse_file_info *fi)
{
    struct dirent *de = NULL;

    log("path=%s\n", path);
    while ((de = readdir ((DIR *) fi->fh)) != NULL)
    {
        struct stat st;
        memset (&st, 0, sizeof (struct stat));
        st.st_ino = de->d_ino;
        st.st_mode = de->d_type << 12;

        if (fill (buf, de->d_name, &st, 0))
            break;
    }

    log("\tReturning 0\n");
    return 0;
}

int
fuse_releasedir (const char *path, struct fuse_file_info *fi)
{
    int retval;

    log("path=%s\n", path);
    retval = closedir ((DIR *) fi->fh) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

int
fuse_access (const char *path, int mode)
{
    char fpath[PATH_MAX];
    int retval;
    fullpath (path, fpath);

    log("path=%s\n", path);
    retval = access (fpath, mode) ? -errno : 0;
    log("\tReturning %d\n", retval);
    return retval;
}

void *
fuse_init (struct fuse_conn_info *conn)
{
#ifdef GENERATE_LOGS
    file_log = fopen("/tmp/transparency.log", "w+");
#endif
    return (fuse_get_context())->private_data;
}

void
fuse_dest(void *conn)
{
#ifdef GENERATE_LOGS
    fclose(file_log);
#endif
}



struct fuse_operations fuse_ops = {
    .getattr = fuse_getattr,
    .readlink = fuse_readlink,
    .mknod = fuse_mknod,
    .mkdir = fuse_mkdir,
    .unlink = fuse_unlink,
    .rmdir = fuse_rmdir,
    .symlink = fuse_symlink,
    .rename = fuse_rename,
    .link = fuse_link,
    .chmod = fuse_chmod,
    .chown = fuse_chown,
    .truncate = fuse_truncate,
    .utime = fuse_utime,
    .open = fuse_open,
    .read = fuse_read,
    .write = fuse_write,
    .statfs = fuse_statfs,
    .release = fuse_release,
    .fsync = fuse_fsync,
    .setxattr = fuse_setxattr,
    .getxattr = fuse_getxattr,
    .listxattr = fuse_listxattr,
    .removexattr = fuse_removexattr,
    .opendir = fuse_opendir,
    .readdir = fuse_readdir,
    .releasedir = fuse_releasedir,
    .access = fuse_access,
    .init = fuse_init,
    .destroy = fuse_dest
};



void
usage ()
{
    printf ("transparency <basedir> <mountpoint>\n");
    exit (0);
}



int
main (int argc, char *argv[])
{
    if (argc < 3)
        usage ();

    if (realpath(argv[1], base) == NULL)
        printf("The base directory path %s is invalid!\n", argv[1]);

    int i = 1;
    for (; i < argc; i++)
        argv[i] = argv[i + 1];
    argc -= 1;

    umask(0);
    int ret = fuse_main (argc, argv, &fuse_ops, base);

    if (ret < 0)
        printf("Error in FUSE main()\n");

    return ret;
}
