/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
  * @file aging.c
  *
  * @brief Implementation Disk Aging
  *
  * @author Praveen Kumar Ramakrishnan <praveenr@andrew.cmu.edu>
  */

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include "aging.h"

static int disk_size = 0;
static int op_count = 0;
static int fileid = 0;

static int max_disk_size = 0;
static int max_op_count = 0;

static randomizer_t create_randomizer;
static randomizer_t delete_randomizer;
static char dirname[MAXLINE];
static char createfile[MAXLINE];
static char deletefile[MAXLINE];
static int num_create = NUM_CREATE;
static int num_delete = NUM_DELETE;
static char filestatsfn[MAXLINE];

int devnull = 0;
unsigned long delete_filesize=0;
unsigned long delete_filecount=0; 
long int response_time_create=0;
long int response_time_delete=0;

void print_usage();
void print_stats();
int read_randomizer_data();
int perform_aging();

int main(int argc, char **argv)
{
    char c;

    strncpy(dirname, OUTPUT_DIR, MAXLINE);
    strncpy(createfile, CREATE_DATA_FILE, MAXLINE);
    strncpy(deletefile, DELETE_DATA_FILE, MAXLINE);
    
    while((c = getopt(argc, argv, "s:n:o:d:c:f:")) != -1) {
        switch(c) {
            case 's':
                max_disk_size = atoi(optarg);
                break;

            case 'n':
                max_op_count = atoi(optarg);
                break;

            case 'o':
                strncpy(dirname, optarg, MAXLINE);
                break;
            
            case 'f':
                strncpy(filestatsfn, optarg, MAXLINE);
                break;

            case 'd':
                strncpy(deletefile, optarg, MAXLINE);
                break;

	    case 'c':
                strncpy(createfile, optarg, MAXLINE);
                break;

            default:
                print_usage();
                return 1;
        }
    }

    if(max_disk_size == 0 || max_op_count == 0) {
        print_usage();
        return 1;
    }

    if(read_randomizer_data() < 0) {
        return 1;
    }

    printf("Randomizer data read successfully!\n");

    if(mkdir(dirname, 0777) < 0) {
        printf("Error! Output directory already exists: %s\n", dirname);
        //return 1;
    }

    printf("Aging started..\n");
    
    srand(RANDOM_SEED);
    devnull = open("/dev/null", O_WRONLY);
    assert(devnull > 0);

    if(perform_aging() < 0) {
        return 1;
    }
    print_stats();
    close(devnull);

    return 0;
}


int read_randomizer_data()
{
    FILE *file = NULL;
    int i, count;
    randomizer_t *r = NULL;

    r = &create_randomizer;
    memset(r, 0, sizeof(randomizer_t));

    file = fopen(createfile, "r");
    if(file == NULL) {
        printf("Error! Create data file not present\n");
        return -1;
    }

    if(fscanf(file, "%d", &count) == EOF) {
        printf("Error! Invalid create data file\n");
        return -1;
    }

    r->count = count;
    r->members = (randomizer_member_t *)calloc(count, 
                                sizeof(randomizer_member_t));
    assert(r->members != NULL);
    
    for(i = 0; i < count; i++) {
        if(fscanf(file, "%d %d %d", 
                  &(r->members[i].weight),
                  &(r->members[i].blocksize),
                  &(r->members[i].blockcount)) == EOF) {
            printf("Error! Invalid create data file\n");
            return -1;
        }
        r->members[i].actualcount = 0;

        r->total_weight += r->members[i].weight;
        Q_INIT_HEAD(&(r->members[i].file_list));
    }

    fclose(file);

    r = &delete_randomizer;
    memset(r, 0, sizeof(randomizer_t));

    file = fopen(deletefile, "r");
    if(file == NULL) {
        printf("Error! Delete data file not present\n");
        return -1;
    }

    if(fscanf(file, "%d", &count) == EOF) {
        printf("Error! Invalid delete data file\n");
        return -1;
    }

    r->count = count;
    r->members = (randomizer_member_t *)calloc(count, 
                                sizeof(randomizer_member_t));
    assert(r->members != NULL);
    
    for(i = 0; i < count; i++) {
        if(fscanf(file, "%d", &(r->members[i].weight)) == EOF) {
            printf("Error! Invalid delete data file\n");
            return -1;
        }
        
        r->total_weight += r->members[i].weight;
        Q_INIT_HEAD(&(r->members[i].file_list));
    }

    fclose(file);

    return 0;
}


char *get_file_name(int id)
{
    static char name[MAXLINE];

    snprintf(name, MAXLINE, "%s/%d", dirname, id);

    return name;
}


randomizer_member_t* get_random_member(randomizer_t *r)
{
   int rnum, i; 
    
    rnum = rand() % (r->total_weight);
    assert(rnum >= 0);
    
    for(i = 0; i < r->count; i++) {
        if(rnum < r->members[i].weight)
            return &(r->members[i]);

        rnum -= r->members[i].weight;
    }

    return NULL;
}


int create_file(randomizer_member_t *rmem)
{
    char *filename = NULL;
    pid_t pid;
    char arg_file[MAXLINE];
    char arg_bs[MAXLINE];
    char arg_count[MAXLINE];
    file_data_t *fd = NULL;
    struct timeval start, end;


    filename = get_file_name(fileid);
    gettimeofday(&start, NULL);
    pid = fork();

    if(pid == 0) {
        dup2(devnull, STDOUT_FILENO);
        dup2(devnull, STDERR_FILENO);
        snprintf(arg_file, MAXLINE, "of=%s", filename);
        snprintf(arg_bs, MAXLINE, "bs=%d", rmem->blocksize);
        snprintf(arg_count, MAXLINE, "count=%d", rmem->blockcount);

        if(execl("/bin/dd", "/bin/dd", "if=/dev/zero", arg_file, arg_bs, arg_count, NULL) < 0) {
            printf("Error! In execv()\n");
            exit(1);
        }
    }
    else if(pid < 0) {
        printf("Error! Cannot fork child\n");
        return -1;
    }
    else {
        waitpid(pid, NULL, 0);
        
        gettimeofday(&end, NULL);
        response_time_create += ((end.tv_sec*1000000 +end.tv_usec) -
                                 (start.tv_sec*1000000 + start.tv_usec));   
        disk_size += (rmem->blockcount * rmem->blocksize);
        printf("Created file: %s, bs: %d, count: %d, disk-size: %d, op-count: %d\n", 
                filename, rmem->blocksize, rmem->blockcount, disk_size, op_count++);
        fd = (file_data_t *)malloc(sizeof(file_data_t));
        assert(fd != NULL);
        memset(fd, 0, sizeof(file_data_t));
        fd->id = fileid;
        fd->blocksize = rmem->blocksize;
        fd->blockcount = rmem->blockcount;
        Q_INIT_ELEM(fd, link);

        Q_INSERT_TAIL(&(delete_randomizer.members[0].file_list),
                      fd, link);
    }

    fileid++;

    return 0;
}


int delete_file(randomizer_member_t *rmem)
{
    file_data_t *fd = NULL;
    char *filename = NULL;
    struct timeval start, end;

    fd = Q_GET_FRONT(&(rmem->file_list));
    assert(fd != NULL);

    Q_REMOVE(&(rmem->file_list), fd, link);

    filename = get_file_name(fd->id);

    gettimeofday(&start, NULL);
    printf("Start Deleting file: %s\n", filename); 
    if(unlink(filename) < 0) {
        printf("Error! Cannot delete file: %s\n", filename);
        return -1;
    }

    gettimeofday(&end, NULL);
    response_time_delete += ((end.tv_sec*1000000 +end.tv_usec) -
                                 (start.tv_sec*1000000 + start.tv_usec));   
    disk_size -= (fd->blockcount * fd->blocksize);
    delete_filesize += (fd->blockcount * fd->blocksize);
    delete_filecount ++;
    printf("Deleted file: %s, disk-size: %d, op-count: %d\n", filename, 
                    disk_size, op_count++);
    
    free(fd);
    return 0;
}


void increase_age()
{
    file_data_queue_t *src= NULL;
    file_data_queue_t *dest = NULL;
    int i;

    for(i = (delete_randomizer.count - 1); i > 0; i--) {
        dest = &(delete_randomizer.members[i].file_list);
        src = &(delete_randomizer.members[i - 1].file_list);

        if(Q_GET_FRONT(dest) != NULL)
            continue;

        memcpy(dest, src, sizeof(file_data_queue_t));

        Q_INIT_HEAD(src);
    }
}


int perform_aging()
{
    int i;
    randomizer_member_t *rmem = NULL;

    while(1) {
        for(i = 0; i < num_create; i++) {
            if(op_count > max_op_count)
                return 0;

            rmem = get_random_member(&create_randomizer);
            assert(rmem != NULL);

            if((disk_size + (rmem->blockcount * rmem->blocksize)) > max_disk_size)
                continue;

            if(create_file(rmem) < 0) {
                return -1;
            }
            rmem->actualcount ++;
        }


        for(i = 0; i < num_delete; i++) {
            if(op_count > max_op_count)
                return 0;

            rmem = get_random_member(&delete_randomizer);
            assert(rmem != NULL);

            if(Q_GET_FRONT(&(rmem->file_list)) == NULL)
                continue;

            if(delete_file(rmem) < 0) {
                return -1;
            }
        }


        if((op_count % INCREASE_AGE_OP_COUNT) == 0) {
            increase_age();
        }
    }

    return 0;
}


void print_usage()
{
    printf("Usage:\n./agedisk -s (disksize) -n (opcount) -c <createfile> -d <deletefile> -o <statsfile>\n");
}

void print_stats()
{
    FILE* filestatsfp=NULL;
    int i = 0;
    time_t create_time;
    randomizer_t *r = NULL;
    r = &create_randomizer;
    
    if (!filestatsfp) {
        filestatsfp = fopen(filestatsfn, "w");
        if (!filestatsfp) {
            fprintf(stderr, "Error creating file: %s\n", filestatsfn);
            return;
        }
    }
    /*print title and timestamp*/
    create_time = time(NULL);
    fprintf(filestatsfp, "#time created: %s", ctime(&create_time));
    fprintf(filestatsfp,"Filesize\tExperimental\tTheoretical\n");
    unsigned long total_bytes = 0;
    unsigned long total_filecount = 0;
    double frequency = 0;
    for(i = 0; i < r->count; i++) {
        total_bytes += r->members[i].blocksize * r->members[i].blockcount * r->members[i].actualcount;
        total_filecount += r->members[i].actualcount;
    }
    for(i = 0; i < r->count; i++) {
        frequency = (double)r->members[i].actualcount / (double) total_filecount;
        fprintf(filestatsfp, "%d\t%f\t%f\n", r->members[i].blocksize * r->members[i].blockcount,
                frequency, 
                (double)r->members[i].weight/(double)r->total_weight);
    }
    fprintf(filestatsfp, "\n#total_bytes created: %lu", total_bytes);
    fprintf(filestatsfp, "\n#total file created: %lu", total_filecount);
    fprintf(filestatsfp, "\n#average file size: %lu", 
            total_bytes/total_filecount);
    fprintf(filestatsfp, "\n#delete file size: %lu", 
            delete_filesize);
    fprintf(filestatsfp, "\n#delete file count: %lu", 
            delete_filecount);
    fprintf(filestatsfp, "\n#create response time: %ld ms", 
            response_time_create/1000);
    fprintf(filestatsfp, "\n#delete response time: %ld ms\n", 
            response_time_delete/1000);
    fclose(filestatsfp);
}
    
